# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PPO vision networks."""

from typing import Any, Callable, Mapping, Sequence, Tuple

from brax.training import distribution
from brax.training import networks
from brax.training import types
import flax
from flax import linen
import jax.numpy as jp


ModuleDef = Any
ActivationFn = Callable[[jp.ndarray], jp.ndarray]
Initializer = Callable[..., Any]


@flax.struct.dataclass
class PPONetworks:
  policy_network: networks.FeedForwardNetwork
  value_network: networks.FeedForwardNetwork
  parametric_action_distribution: distribution.ParametricDistribution


def make_ppo_networks_vision(
    observation_size: Mapping[str, Tuple[int, ...]],
    action_size: int,
    preprocess_observations_fn: types.PreprocessObservationFn = types.identity_observation_preprocessor,
    policy_hidden_layer_sizes: Sequence[int] = (256, 256),
    value_hidden_layer_sizes: Sequence[int] = (256, 256),
    activation: ActivationFn = linen.swish,
    normalise_channels: bool = False,
    policy_obs_key: str = "",
    value_obs_key: str = "",
) -> PPONetworks:
  """Make Vision PPO networks with preprocessor."""

  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=action_size
  )

  policy_network = networks.make_policy_network_vision(
      observation_size=observation_size,
      output_size=parametric_action_distribution.param_size,
      preprocess_observations_fn=preprocess_observations_fn,
      activation=activation,
      hidden_layer_sizes=policy_hidden_layer_sizes,
      state_obs_key=policy_obs_key,
      normalise_channels=normalise_channels,
  )

  value_network = networks.make_value_network_vision(
      observation_size=observation_size,
      preprocess_observations_fn=preprocess_observations_fn,
      activation=activation,
      hidden_layer_sizes=value_hidden_layer_sizes,
      state_obs_key=value_obs_key,
      normalise_channels=normalise_channels,
  )

  return PPONetworks(
      policy_network=policy_network,
      value_network=value_network,
      parametric_action_distribution=parametric_action_distribution,
  )



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Proximal policy optimization training.

See: https://arxiv.org/pdf/1707.06347.pdf
"""

import functools
import time
from typing import Any, Callable, Mapping, Optional, Tuple, Union

from absl import logging
from brax import base
from brax import envs
from brax.training import acting
from brax.training import gradients
from brax.training import logger as metric_logger
from brax.training import pmap
from brax.training import types
from brax.training.acme import running_statistics
from brax.training.acme import specs
from brax.training.agents.ppo import checkpoint
from brax.training.agents.ppo import losses as ppo_losses
from brax.training.agents.ppo import networks as ppo_networks
from brax.training.types import Params
from brax.training.types import PRNGKey
import flax
import jax
import jax.numpy as jnp
import numpy as np
import optax


InferenceParams = Tuple[running_statistics.NestedMeanStd, Params]
Metrics = types.Metrics

_PMAP_AXIS_NAME = 'i'


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""

  optimizer_state: optax.OptState
  params: ppo_losses.PPONetworkParams
  normalizer_params: running_statistics.RunningStatisticsState
  env_steps: types.UInt64


def _unpmap(v):
  return jax.tree_util.tree_map(lambda x: x[0], v)


def _strip_weak_type(tree):
  # brax user code is sometimes ambiguous about weak_type.  in order to
  # avoid extra jit recompilations we strip all weak types from user input
  def f(leaf):
    leaf = jnp.asarray(leaf)
    return jnp.astype(leaf, leaf.dtype)

  return jax.tree_util.tree_map(f, tree)


def _validate_madrona_args(
    madrona_backend: bool,
    num_envs: int,
    num_eval_envs: int,
    action_repeat: int,
    eval_env: Optional[envs.Env] = None,
):
  """Validates arguments for Madrona-MJX."""
  if madrona_backend:
    if eval_env:
      raise ValueError("Madrona-MJX doesn't support multiple env instances")
    if num_eval_envs != num_envs:
      raise ValueError('Madrona-MJX requires a fixed batch size')
    if action_repeat != 1:
      raise ValueError(
          "Implement action_repeat using PipelineEnv's _n_frames to avoid"
          ' unnecessary rendering!'
      )


def _maybe_wrap_env(
    env: envs.Env,
    wrap_env: bool,
    num_envs: int,
    episode_length: Optional[int],
    action_repeat: int,
    device_count: int,
    key_env: PRNGKey,
    wrap_env_fn: Optional[Callable[[Any], Any]] = None,
    randomization_fn: Optional[
        Callable[[base.System, jnp.ndarray], Tuple[base.System, base.System]]
    ] = None,
):
  """Wraps the environment for training/eval if wrap_env is True."""
  if not wrap_env:
    return env
  if episode_length is None:
    raise ValueError('episode_length must be specified in ppo.train')
  v_randomization_fn = None
  if randomization_fn is not None:
    randomization_batch_size = num_envs // device_count
    # all devices gets the same randomization rng
    randomization_rng = jax.random.split(key_env, randomization_batch_size)
    v_randomization_fn = functools.partial(
        randomization_fn, rng=randomization_rng
    )
  if wrap_env_fn is not None:
    wrap_for_training = wrap_env_fn
  else:
    wrap_for_training = envs.training.wrap
  env = wrap_for_training(
      env,
      episode_length=episode_length,
      action_repeat=action_repeat,
      randomization_fn=v_randomization_fn,
  )  # pytype: disable=wrong-keyword-args
  return env


def _random_translate_pixels(
    obs: Mapping[str, jax.Array], key: PRNGKey
) -> Mapping[str, jax.Array]:
  """Apply random translations to B x T x ... pixel observations.

  The same shift is applied across the unroll_length (T) dimension.

  Args:
    obs: a dictionary of observations
    key: a PRNGKey

  Returns:
    A dictionary of observations with translated pixels
  """

  @jax.vmap
  def rt_all_views(
      ub_obs: Mapping[str, jax.Array], key: PRNGKey
  ) -> Mapping[str, jax.Array]:
    # Expects dictionary of unbatched observations.
    def rt_view(
        img: jax.Array, padding: int, key: PRNGKey
    ) -> jax.Array:  # TxHxWxC
      # Randomly translates a set of pixel inputs.
      # Adapted from
      # https://github.com/ikostrikov/jaxrl/blob/main/jaxrl/agents/drq/augmentations.py
      crop_from = jax.random.randint(key, (2,), 0, 2 * padding + 1)
      zero = jnp.zeros((1,), dtype=jnp.int32)
      crop_from = jnp.concatenate([zero, crop_from, zero])
      padded_img = jnp.pad(
          img,
          ((0, 0), (padding, padding), (padding, padding), (0, 0)),
          mode='edge',
      )
      return jax.lax.dynamic_slice(padded_img, crop_from, img.shape)

    out = {}
    for k_view, v_view in ub_obs.items():
      if k_view.startswith('pixels/'):
        key, key_shift = jax.random.split(key)
        out[k_view] = rt_view(v_view, 4, key_shift)
    return {**ub_obs, **out}

  bdim = next(iter(obs.items()), None)[1].shape[0]
  keys = jax.random.split(key, bdim)
  obs = rt_all_views(obs, keys)
  return obs


def _remove_pixels(
    obs: Union[jnp.ndarray, Mapping[str, jax.Array]],
) -> Union[jnp.ndarray, Mapping[str, jax.Array]]:
  """Removes pixel observations from the observation dict."""
  if not isinstance(obs, Mapping):
    return obs
  return {k: v for k, v in obs.items() if not k.startswith('pixels/')}


def train(
    environment: envs.Env,
    num_timesteps: int,
    max_devices_per_host: Optional[int] = None,
    # high-level control flow
    wrap_env: bool = True,
    madrona_backend: bool = False,
    augment_pixels: bool = False,
    # environment wrapper
    num_envs: int = 1,
    episode_length: Optional[int] = None,
    action_repeat: int = 1,
    wrap_env_fn: Optional[Callable[[Any], Any]] = None,
    randomization_fn: Optional[
        Callable[[base.System, jnp.ndarray], Tuple[base.System, base.System]]
    ] = None,
    # ppo params
    learning_rate: float = 1e-4,
    entropy_cost: float = 1e-4,
    discounting: float = 0.9,
    unroll_length: int = 10,
    batch_size: int = 32,
    num_minibatches: int = 16,
    num_updates_per_batch: int = 2,
    num_resets_per_eval: int = 0,
    normalize_observations: bool = False,
    reward_scaling: float = 1.0,
    clipping_epsilon: float = 0.3,
    gae_lambda: float = 0.95,
    max_grad_norm: Optional[float] = None,
    normalize_advantage: bool = True,
    network_factory: types.NetworkFactory[
        ppo_networks.PPONetworks
    ] = ppo_networks.make_ppo_networks,
    seed: int = 0,
    use_pmap_on_reset: bool = True,
    # eval
    num_evals: int = 1,
    eval_env: Optional[envs.Env] = None,
    num_eval_envs: int = 128,
    deterministic_eval: bool = False,
    # training metrics
    log_training_metrics: bool = False,
    training_metrics_steps: Optional[int] = None,
    # callbacks
    progress_fn: Callable[[int, Metrics], None] = lambda *args: None,
    policy_params_fn: Callable[..., None] = lambda *args: None,
    # checkpointing
    save_checkpoint_path: Optional[str] = None,
    restore_checkpoint_path: Optional[str] = None,
    restore_params: Optional[Any] = None,
    restore_value_fn: bool = True,
    run_evals: bool = True,
):
  """PPO training.

  Args:
    environment: the environment to train
    num_timesteps: the total number of environment steps to use during training
    max_devices_per_host: maximum number of chips to use per host process
    wrap_env: If True, wrap the environment for training. Otherwise use the
      environment as is.
    madrona_backend: whether to use Madrona backend for training
    augment_pixels: whether to add image augmentation to pixel inputs
    num_envs: the number of parallel environments to use for rollouts
      NOTE: `num_envs` must be divisible by the total number of chips since each
        chip gets `num_envs // total_number_of_chips` environments to roll out
      NOTE: `batch_size * num_minibatches` must be divisible by `num_envs` since
        data generated by `num_envs` parallel envs gets used for gradient
        updates over `num_minibatches` of data, where each minibatch has a
        leading dimension of `batch_size`
    episode_length: the length of an environment episode
    action_repeat: the number of timesteps to repeat an action
    wrap_env_fn: a custom function that wraps the environment for training. If
      not specified, the environment is wrapped with the default training
      wrapper.
    randomization_fn: a user-defined callback function that generates randomized
      environments
    learning_rate: learning rate for ppo loss
    entropy_cost: entropy reward for ppo loss, higher values increase entropy of
      the policy
    discounting: discounting rate
    unroll_length: the number of timesteps to unroll in each environment. The
      PPO loss is computed over `unroll_length` timesteps
    batch_size: the batch size for each minibatch SGD step
    num_minibatches: the number of times to run the SGD step, each with a
      different minibatch with leading dimension of `batch_size`
    num_updates_per_batch: the number of times to run the gradient update over
      all minibatches before doing a new environment rollout
    num_resets_per_eval: the number of environment resets to run between each
      eval. The environment resets occur on the host
    normalize_observations: whether to normalize observations
    reward_scaling: float scaling for reward
    clipping_epsilon: clipping epsilon for PPO loss
    gae_lambda: General advantage estimation lambda
    max_grad_norm: gradient clipping norm value. If None, no clipping is done
    normalize_advantage: whether to normalize advantage estimate
    network_factory: function that generates networks for policy and value
      functions
    seed: random seed
    num_evals: the number of evals to run during the entire training run.
      Increasing the number of evals increases total training time
    eval_env: an optional environment for eval only, defaults to `environment`
    num_eval_envs: the number of envs to use for evluation. Each env will run 1
      episode, and all envs run in parallel during eval.
    deterministic_eval: whether to run the eval with a deterministic policy
    log_training_metrics: whether to log training metrics and callback to
      progress_fn
    training_metrics_steps: the number of environment steps between logging
      training metrics
    progress_fn: a user-defined callback function for reporting/plotting metrics
    policy_params_fn: a user-defined callback function that can be used for
      saving custom policy checkpoints or creating policy rollouts and videos
    save_checkpoint_path: the path used to save checkpoints. If None, no
      checkpoints are saved.
    restore_checkpoint_path: the path used to restore previous model params
    restore_params: raw network parameters to restore the TrainingState from.
      These override `restore_checkpoint_path`. These paramaters can be obtained
      from the return values of ppo.train().
    restore_value_fn: whether to restore the value function from the checkpoint
      or use a random initialization
    run_evals: if True, use the evaluator num_eval times to collect distinct
      eval rollouts. If False, num_eval_envs and eval_env are ignored.
      progress_fn is then expected to use training_metrics.
    use_pmap_on_reset: default to True. if True, use pmap instead of vmap for
      env.reset across devices.

  Returns:
    Tuple of (make_policy function, network params, metrics)
  """
  assert batch_size * num_minibatches % num_envs == 0
  _validate_madrona_args(
      madrona_backend, num_envs, num_eval_envs, action_repeat, eval_env
  )

  xt = time.time()

  process_count = jax.process_count()
  process_id = jax.process_index()
  local_device_count = jax.local_device_count()
  local_devices_to_use = local_device_count
  if max_devices_per_host:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  logging.info(
      'Device count: %d, process count: %d (id %d), local device count: %d, '
      'devices to be used count: %d',
      jax.device_count(),
      process_count,
      process_id,
      local_device_count,
      local_devices_to_use,
  )
  device_count = local_devices_to_use * process_count

  # The number of environment steps executed for every training step.
  env_step_per_training_step = (
      batch_size * unroll_length * num_minibatches * action_repeat
  )
  num_evals_after_init = max(num_evals - 1, 1)
  # The number of training_step calls per training_epoch call.
  # equals to ceil(num_timesteps / (num_evals * env_step_per_training_step *
  #                                 num_resets_per_eval))
  num_training_steps_per_epoch = np.ceil(
      num_timesteps
      / (
          num_evals_after_init
          * env_step_per_training_step
          * max(num_resets_per_eval, 1)
      )
  ).astype(int)

  key = jax.random.PRNGKey(seed)
  global_key, local_key = jax.random.split(key)
  del key
  local_key = jax.random.fold_in(local_key, process_id)
  local_key, key_env, eval_key = jax.random.split(local_key, 3)
  # key_networks should be global, so that networks are initialized the same
  # way for different processes.
  key_policy, key_value = jax.random.split(global_key)
  del global_key

  assert num_envs % device_count == 0

  env = _maybe_wrap_env(
      environment,
      wrap_env,
      num_envs,
      episode_length,
      action_repeat,
      device_count,
      key_env,
      wrap_env_fn,
      randomization_fn,
  )

  if local_devices_to_use > 1 or use_pmap_on_reset:
    reset_fn = jax.pmap(env.reset, axis_name=_PMAP_AXIS_NAME)
  else:
    reset_fn = jax.jit(jax.vmap(env.reset))

  key_envs = jax.random.split(key_env, num_envs // process_count)
  key_envs = jnp.reshape(
      key_envs, (local_devices_to_use, -1) + key_envs.shape[1:]
  )
  env_state = reset_fn(key_envs)
  # Discard the batch axes over devices and envs.
  obs_shape = jax.tree_util.tree_map(lambda x: x.shape[2:], env_state.obs)

  normalize = lambda x, y: x
  if normalize_observations:
    normalize = running_statistics.normalize
  ppo_network = network_factory(
      obs_shape, env.action_size, preprocess_observations_fn=normalize
  )
  make_policy = ppo_networks.make_inference_fn(ppo_network)

  optimizer = optax.adam(learning_rate=learning_rate)
  if max_grad_norm is not None:
    # TODO(btaba): Move gradient clipping to `training/gradients.py`.
    optimizer = optax.chain(
        optax.clip_by_global_norm(max_grad_norm),
        optax.adam(learning_rate=learning_rate),
    )

  loss_fn = functools.partial(
      ppo_losses.compute_ppo_loss,
      ppo_network=ppo_network,
      entropy_cost=entropy_cost,
      discounting=discounting,
      reward_scaling=reward_scaling,
      gae_lambda=gae_lambda,
      clipping_epsilon=clipping_epsilon,
      normalize_advantage=normalize_advantage,
  )

  gradient_update_fn = gradients.gradient_update_fn(
      loss_fn, optimizer, pmap_axis_name=_PMAP_AXIS_NAME, has_aux=True
  )

  metrics_aggregator = metric_logger.EpisodeMetricsLogger(
      steps_between_logging=training_metrics_steps
      or env_step_per_training_step,
      progress_fn=progress_fn,
  )

  def minibatch_step(
      carry,
      data: types.Transition,
      normalizer_params: running_statistics.RunningStatisticsState,
  ):
    optimizer_state, params, key = carry
    key, key_loss = jax.random.split(key)
    (_, metrics), params, optimizer_state = gradient_update_fn(
        params,
        normalizer_params,
        data,
        key_loss,
        optimizer_state=optimizer_state,
    )

    return (optimizer_state, params, key), metrics

  def sgd_step(
      carry,
      unused_t,
      data: types.Transition,
      normalizer_params: running_statistics.RunningStatisticsState,
  ):
    optimizer_state, params, key = carry
    key, key_perm, key_grad = jax.random.split(key, 3)

    if augment_pixels:
      key, key_rt = jax.random.split(key)
      r_translate = functools.partial(_random_translate_pixels, key=key_rt)
      data = types.Transition(
          observation=r_translate(data.observation),
          action=data.action,
          reward=data.reward,
          discount=data.discount,
          next_observation=r_translate(data.next_observation),
          extras=data.extras,
      )

    def convert_data(x: jnp.ndarray):
      x = jax.random.permutation(key_perm, x)
      x = jnp.reshape(x, (num_minibatches, -1) + x.shape[1:])
      return x

    shuffled_data = jax.tree_util.tree_map(convert_data, data)
    (optimizer_state, params, _), metrics = jax.lax.scan(
        functools.partial(minibatch_step, normalizer_params=normalizer_params),
        (optimizer_state, params, key_grad),
        shuffled_data,
        length=num_minibatches,
    )
    return (optimizer_state, params, key), metrics

  def training_step(
      carry: Tuple[TrainingState, envs.State, PRNGKey], unused_t
  ) -> Tuple[Tuple[TrainingState, envs.State, PRNGKey], Metrics]:
    training_state, state, key = carry
    key_sgd, key_generate_unroll, new_key = jax.random.split(key, 3)

    policy = make_policy((
        training_state.normalizer_params,
        training_state.params.policy,
        training_state.params.value,
    ))

    def f(carry, unused_t):
      current_state, current_key = carry
      current_key, next_key = jax.random.split(current_key)
      next_state, data = acting.generate_unroll(
          env,
          current_state,
          policy,
          current_key,
          unroll_length,
          extra_fields=('truncation', 'episode_metrics', 'episode_done'),
      )
      return (next_state, next_key), data

    (state, _), data = jax.lax.scan(
        f,
        (state, key_generate_unroll),
        (),
        length=batch_size * num_minibatches // num_envs,
    )
    # Have leading dimensions (batch_size * num_minibatches, unroll_length)
    data = jax.tree_util.tree_map(lambda x: jnp.swapaxes(x, 1, 2), data)
    data = jax.tree_util.tree_map(
        lambda x: jnp.reshape(x, (-1,) + x.shape[2:]), data
    )
    assert data.discount.shape[1:] == (unroll_length,)

    if log_training_metrics:  # log unroll metrics
      jax.debug.callback(
          metrics_aggregator.update_episode_metrics,
          data.extras['state_extras']['episode_metrics'],
          data.extras['state_extras']['episode_done'],
      )

    # Update normalization params and normalize observations.
    normalizer_params = running_statistics.update(
        training_state.normalizer_params,
        _remove_pixels(data.observation),
        pmap_axis_name=_PMAP_AXIS_NAME,
    )

    (optimizer_state, params, _), metrics = jax.lax.scan(
        functools.partial(
            sgd_step, data=data, normalizer_params=normalizer_params
        ),
        (training_state.optimizer_state, training_state.params, key_sgd),
        (),
        length=num_updates_per_batch,
    )

    new_training_state = TrainingState(
        optimizer_state=optimizer_state,
        params=params,
        normalizer_params=normalizer_params,
        env_steps=training_state.env_steps + env_step_per_training_step,
    )
    return (new_training_state, state, new_key), metrics

  def training_epoch(
      training_state: TrainingState, state: envs.State, key: PRNGKey
  ) -> Tuple[TrainingState, envs.State, Metrics]:
    (training_state, state, _), loss_metrics = jax.lax.scan(
        training_step,
        (training_state, state, key),
        (),
        length=num_training_steps_per_epoch,
    )
    loss_metrics = jax.tree_util.tree_map(jnp.mean, loss_metrics)
    return training_state, state, loss_metrics

  training_epoch = jax.pmap(training_epoch, axis_name=_PMAP_AXIS_NAME)

  # Note that this is NOT a pure jittable method.
  def training_epoch_with_timing(
      training_state: TrainingState, env_state: envs.State, key: PRNGKey
  ) -> Tuple[TrainingState, envs.State, Metrics]:
    nonlocal training_walltime
    t = time.time()
    training_state, env_state = _strip_weak_type((training_state, env_state))
    result = training_epoch(training_state, env_state, key)
    training_state, env_state, metrics = _strip_weak_type(result)

    metrics = jax.tree_util.tree_map(jnp.mean, metrics)
    jax.tree_util.tree_map(lambda x: x.block_until_ready(), metrics)

    epoch_training_time = time.time() - t
    training_walltime += epoch_training_time
    sps = (
        num_training_steps_per_epoch
        * env_step_per_training_step
        * max(num_resets_per_eval, 1)
    ) / epoch_training_time
    metrics = {
        'training/sps': sps,
        'training/walltime': training_walltime,
        **{f'training/{name}': value for name, value in metrics.items()},
    }
    return training_state, env_state, metrics  # pytype: disable=bad-return-type  # py311-upgrade

  # Initialize model params and training state.
  init_params = ppo_losses.PPONetworkParams(
      policy=ppo_network.policy_network.init(key_policy),
      value=ppo_network.value_network.init(key_value),
  )

  obs_shape = jax.tree_util.tree_map(
      lambda x: specs.Array(x.shape[-1:], jnp.dtype('float32')), env_state.obs
  )
  training_state = TrainingState(  # pytype: disable=wrong-arg-types  # jax-ndarray
      optimizer_state=optimizer.init(init_params),  # pytype: disable=wrong-arg-types  # numpy-scalars
      params=init_params,
      normalizer_params=running_statistics.init_state(
          _remove_pixels(obs_shape)
      ),
      env_steps=types.UInt64(hi=0, lo=0),
  )

  if restore_checkpoint_path is not None:
    params = checkpoint.load(restore_checkpoint_path)
    value_params = params[2] if restore_value_fn else init_params.value
    training_state = training_state.replace(
        normalizer_params=params[0],
        params=training_state.params.replace(
            policy=params[1], value=value_params
        ),
    )

  if restore_params is not None:
    logging.info('Restoring TrainingState from `restore_params`.')
    value_params = restore_params[2] if restore_value_fn else init_params.value
    training_state = training_state.replace(
        normalizer_params=restore_params[0],
        params=training_state.params.replace(
            policy=restore_params[1], value=value_params
        ),
    )

  if num_timesteps == 0:
    return (
        make_policy,
        (
            training_state.normalizer_params,
            training_state.params.policy,
            training_state.params.value,
        ),
        {},
    )

  training_state = jax.device_put_replicated(
      training_state, jax.local_devices()[:local_devices_to_use]
  )

  eval_env = _maybe_wrap_env(
      eval_env or environment,
      wrap_env,
      num_eval_envs,
      episode_length,
      action_repeat,
      device_count=1,  # eval on the host only
      key_env=eval_key,
      wrap_env_fn=wrap_env_fn,
      randomization_fn=randomization_fn,
  )
  evaluator = acting.Evaluator(
      eval_env,
      functools.partial(make_policy, deterministic=deterministic_eval),
      num_eval_envs=num_eval_envs,
      episode_length=episode_length,
      action_repeat=action_repeat,
      key=eval_key,
  )

  training_metrics = {}
  training_walltime = 0
  current_step = 0

  # Run initial eval
  metrics = {}
  if process_id == 0 and num_evals > 1 and run_evals:
    metrics = evaluator.run_evaluation(
        _unpmap((
            training_state.normalizer_params,
            training_state.params.policy,
            training_state.params.value,
        )),
        training_metrics={},
    )
    logging.info(metrics)
    progress_fn(0, metrics)

  # Run initial policy_params_fn.
  params = _unpmap((
      training_state.normalizer_params,
      training_state.params.policy,
      training_state.params.value,
  ))
  policy_params_fn(current_step, make_policy, params)

  for it in range(num_evals_after_init):
    logging.info('starting iteration %s %s', it, time.time() - xt)

    for _ in range(max(num_resets_per_eval, 1)):
      # optimization
      epoch_key, local_key = jax.random.split(local_key)
      epoch_keys = jax.random.split(epoch_key, local_devices_to_use)
      (training_state, env_state, training_metrics) = (
          training_epoch_with_timing(training_state, env_state, epoch_keys)
      )
      current_step = int(_unpmap(training_state.env_steps))

      key_envs = jax.vmap(
          lambda x, s: jax.random.split(x[0], s), in_axes=(0, None)
      )(key_envs, key_envs.shape[1])
      # TODO(brax-team): move extra reset logic to the AutoResetWrapper.
      env_state = reset_fn(key_envs) if num_resets_per_eval > 0 else env_state

    if process_id != 0:
      continue

    # Process id == 0.
    params = _unpmap((
        training_state.normalizer_params,
        training_state.params.policy,
        training_state.params.value,
    ))

    policy_params_fn(current_step, make_policy, params)

    if save_checkpoint_path is not None:
      ckpt_config = checkpoint.network_config(
          observation_size=obs_shape,
          action_size=env.action_size,
          normalize_observations=normalize_observations,
          network_factory=network_factory,
      )
      checkpoint.save(
          save_checkpoint_path, current_step, params, ckpt_config
      )

    if num_evals > 0:
      metrics = training_metrics
      if run_evals:
        metrics = evaluator.run_evaluation(
            params,
            training_metrics,
        )
      logging.info(metrics)
      progress_fn(current_step, metrics)

  total_steps = current_step
  if not total_steps >= num_timesteps:
    raise AssertionError(
        f'Total steps {total_steps} is less than `num_timesteps`='
        f' {num_timesteps}.'
    )

  # If there was no mistakes the training_state should still be identical on all
  # devices.
  pmap.assert_is_replicated(training_state)
  params = _unpmap((
      training_state.normalizer_params,
      training_state.params.policy,
      training_state.params.value,
  ))
  logging.info('total steps: %s', total_steps)
  pmap.synchronize_hosts()
  return (make_policy, params, metrics)



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brax training types."""

from absl.testing import absltest
from brax.training import types
import jax
from jax import numpy as jp

UInt64 = types.UInt64


class TestUInt64(absltest.TestCase):
  """Tests for UInt64 class and its operations."""

  def test_add_uint64_basic(self):
    """Tests basic addition of two UInt64 numbers."""
    a = UInt64(hi=jp.array(0), lo=jp.array(1))
    b = UInt64(hi=jp.array(0), lo=jp.array(2))
    expected = UInt64(hi=jp.array(0), lo=jp.array(3))
    self.assertEqual(a + b, expected)

  def test_add_uint64_with_carry(self):
    """Tests addition of two UInt64 numbers with carry-over."""
    max_uint32 = 2**32 - 1
    a = UInt64(hi=jp.array(0), lo=jp.array(max_uint32, dtype=jp.uint32))
    b = UInt64(hi=jp.array(0), lo=jp.array(1))
    expected = UInt64(hi=jp.array(1), lo=jp.array(0))
    self.assertEqual(a + b, expected)

  def test_add_uint64_large_numbers(self):
    """Tests addition with larger numbers that will result in non-zero hi."""
    a = UInt64(hi=jp.array(1), lo=jp.array(0))
    b = UInt64(hi=jp.array(2), lo=jp.array(0))
    expected = UInt64(hi=jp.array(3), lo=jp.array(0))
    self.assertEqual(a + b, expected)

  def test_add_uint64_max_values(self):
    """Tests addition of the maximum possible UInt64 values."""
    max_uint32 = jp.array(2**32 - 1, dtype=jp.uint32)
    a = UInt64(hi=max_uint32, lo=max_uint32)
    b = UInt64(hi=max_uint32, lo=max_uint32)
    expected_hi = jp.array(2**32 - 1, dtype=jp.uint32)
    expected_lo = jp.array(2**32 - 2, dtype=jp.uint32)
    expected = UInt64(hi=expected_hi, lo=expected_lo)
    self.assertEqual(a + b, expected)

  def test_uint64_add_different_type(self):
    """Tests adding UInt64 with a non-UInt64 type."""
    a = UInt64(hi=jp.array(0), lo=jp.array(1))
    self.assertEqual(a + 1, UInt64(hi=jp.array(0), lo=jp.array(2)))

  def test_uint64_repr(self):
    """Tests the __repr__ method of UInt64."""
    a = UInt64(hi=jp.array(1), lo=jp.array(20))
    self.assertEqual(repr(a), f"{UInt64(hi=jp.array(1), lo=jp.array(20))}")

  def test_add_uint64_jit(self):
    """Tests UInt64 addition under jit."""

    @jax.jit
    def add_uint64(a, b):
      return a + b

    a = UInt64(hi=jp.array(1), lo=jp.array(2**32 - 1, dtype=jp.uint32))
    b = UInt64(hi=jp.array(2), lo=jp.array(1))
    result = add_uint64(a, b)
    expected = UInt64(hi=jp.array(4), lo=jp.array(0))
    self.assertEqual(result, expected)

  def test_uint64_to_numpy(self):
    """Tests the to_numpy method of UInt64."""
    a = UInt64(hi=jp.array(1), lo=jp.array(5120))
    self.assertEqual(a.to_numpy(), (1 << 32) + 5120)

  def test_uint64_int(self):
    """Tests the int method of UInt64."""
    a = UInt64(hi=jp.array(1), lo=jp.array(20))
    self.assertEqual(int(a), (1 << 32) + 20)


if __name__ == "__main__":
  absltest.main()



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BC networks."""

from typing import Any, Callable, Mapping, Optional, Sequence, Tuple

from brax.training import distribution
from brax.training import networks
from brax.training import types
from brax.training.types import PRNGKey
import flax
from flax import linen
import jax.numpy as jp


@flax.struct.dataclass
class BCNetworks:
  policy_network: networks.FeedForwardNetwork
  parametric_action_distribution: distribution.ParametricDistribution


BCInferenceFn = Callable[
    [types.Observation, PRNGKey], Tuple[jp.ndarray, Mapping[str, Any]]
]


def make_inference_fn(bc_networks: BCNetworks):
  """Creates params and inference function for the PPO agent."""

  def make_policy(
      params: types.Params,
      *,
      deterministic: bool = True,
      tanh_squash: bool = False,
  ) -> BCInferenceFn:
    """Keeping unused deterministic and key_sample for API compatibility.

    (BC inference is always deterministic)
    """
    policy_network = bc_networks.policy_network
    parametric_action_distribution = bc_networks.parametric_action_distribution

    def policy(
        observations: types.Observation, key_sample: Optional[PRNGKey]
    ) -> Tuple[types.Action, types.Extra]:
      param_subset = (params[0], params[1])  # normalizer and policy params
      logits = policy_network.apply(*param_subset, observations)
      action_dist = parametric_action_distribution.create_dist(logits)
      act = (
          parametric_action_distribution.mode(logits)
          if tanh_squash
          else action_dist.loc
      )
      metainfo = {'loc': action_dist.loc, 'scale': action_dist.scale}
      return act, metainfo

    return policy

  return make_policy


def make_bc_networks(
    observation_size: Mapping[str, Tuple[int, ...]],
    action_size: int,
    preprocess_observations_fn: types.PreprocessObservationFn = types.identity_observation_preprocessor,
    *,
    policy_hidden_layer_sizes: Sequence[int] = (32,) * 4,
    activation: networks.ActivationFn = linen.swish,
    policy_obs_key: str = '',
    vision: bool = False,
    latent_vision: bool = False,
) -> BCNetworks:
  """Make BC networks with preprocessor.

  Note that vision = True assumes Frozen Encoder.
  """
  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=action_size
  )

  if latent_vision:
    make_policy_network = networks.make_policy_network_latents
  elif vision:
    make_policy_network = networks.make_policy_network_vision
  else:
    make_policy_network = networks.make_policy_network

  policy_network = make_policy_network(
      parametric_action_distribution.param_size,
      observation_size,
      preprocess_observations_fn=preprocess_observations_fn,
      hidden_layer_sizes=policy_hidden_layer_sizes,
      activation=activation,
      obs_key=policy_obs_key,
  )

  return BCNetworks(
      policy_network=policy_network,
      parametric_action_distribution=parametric_action_distribution,
  )



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RL training with an environment running entirely on an accelerator."""

import functools
import json
import os

os.environ['XLA_FLAGS'] = '--xla_gpu_graph_min_graph_size=1'

from absl import app
from absl import flags
from absl import logging
from brax import envs
from brax.io import html
from brax.io import metrics
from brax.io import model
from brax.training.agents.apg import train as apg
from brax.training.agents.ars import train as ars
from brax.training.agents.es import train as es
from brax.training.agents.ppo import networks as ppo_networks
from brax.training.agents.ppo import train as ppo
from brax.training.agents.sac import networks as sac_networks
from brax.training.agents.sac import train as sac
from etils import epath
import jax
import mediapy as media

FLAGS = flags.FLAGS
_LEARNER = flags.DEFINE_enum(
    'learner',
    'ppo',
    ['ppo', 'apg', 'es', 'sac', 'ars'],
    'Which algorithm to run.',
)
_ENV = flags.DEFINE_string('env', 'ant', 'Name of environment to train.')

_BACKEND = flags.DEFINE_enum(
    'backend',
    'mjx',
    ['mjx', 'spring', 'generalized', 'positional'],
    'The physics backend to use.',
)
_TOTAL_ENV_STEPS = flags.DEFINE_integer(
    'total_env_steps', 50000000, 'Number of env steps to run training for.'
)
_NUM_EVALS = flags.DEFINE_integer(
    'num_evals', 10, 'How many times to run an eval.'
)
_SEED = flags.DEFINE_integer('seed', 0, 'Random seed.')
_NUM_ENVS = flags.DEFINE_integer(
    'num_envs', 4, 'Number of envs to run in parallel.'
)
_ACTION_REPEAT = flags.DEFINE_integer('action_repeat', 1, 'Action repeat.')
_UNROLL_LENGTH = flags.DEFINE_integer('unroll_length', 30, 'Unroll length.')
_BATCH_SIZE = flags.DEFINE_integer('batch_size', 4, 'Batch size.')
_NUM_MINIBATCHES = flags.DEFINE_integer('num_minibatches', 1, 'Number')
_NUM_UPDATES_PER_BATCH = flags.DEFINE_integer(
    'num_updates_per_batch',
    1,
    'Number of times to reuse each transition for gradient computation.',
)
_REWARD_SCALING = flags.DEFINE_float('reward_scaling', 10.0, 'Reward scale.')
_ENTROPY_COST = flags.DEFINE_float('entropy_cost', 3e-4, 'Entropy cost.')
_EPISODE_LENGTH = flags.DEFINE_integer(
    'episode_length', 1000, 'Episode length.'
)
_DISCOUNTING = flags.DEFINE_float('discounting', 0.99, 'Discounting.')
_LEARNING_RATE = flags.DEFINE_float('learning_rate', 5e-4, 'Learning rate.')
_MAX_GRADIENT_NORM = flags.DEFINE_float(
    'max_gradient_norm', 1e9, 'Maximal norm of a gradient update.'
)
_LOGDIR = flags.DEFINE_string('logdir', '', 'Logdir.')
_NORMALIZE_OBSERVATIONS = flags.DEFINE_bool(
    'normalize_observations',
    True,
    'Whether to apply observation normalization.',
)
_MAX_DEVICES_PER_HOST = flags.DEFINE_integer(
    'max_devices_per_host',
    None,
    'Maximum number of devices to use per host. If None, '
    'defaults to use as much as it can.',
)
_NUM_VIDEOS = flags.DEFINE_integer(
    'num_videos', 1, 'Number of videos to record after training.'
)
# Evolution Strategy related flags
_POPULATION_SIZE = flags.DEFINE_integer(
    'population_size',
    1,
    'Number of environments in ES. The actual number is 2x '
    'larger (used for antithetic sampling.',
)
_PERTURBATION_STD = flags.DEFINE_float(
    'perturbation_std', 0.1, 'Std of a random noise added by ES.'
)
_FITNESS_SHAPING = flags.DEFINE_enum(
    'fitness_shaping',
    'original',
    ['original', 'centered_rank', 'wierstra'],
    'Defines a type of fitness shaping to apply.',
)
_CENTER_FITNESS = flags.DEFINE_bool(
    'center_fitness', False, 'Whether to normalize fitness after the shaping.'
)
_L2COEFF = flags.DEFINE_float(
    'l2coeff', 0, 'L2 regularization coefficient for model params.'
)
# SAC hps.
_MIN_REPLAY_SIZE = flags.DEFINE_integer(
    'min_replay_size',
    8192,
    'Minimal replay buffer size before the training starts.',
)
_MAX_REPLAY_SIZE = flags.DEFINE_integer(
    'max_replay_size', 1048576, 'Maximal replay buffer size.'
)
_GRAD_UPDATES_PER_STEP = flags.DEFINE_integer(
    'grad_updates_per_step',
    1,
    'How many SAC gradient updates to run per one step in the environment.',
)
_Q_NETWORK_LAYER_NORM = flags.DEFINE_bool(
    'q_network_layer_norm', False, 'Critic network layer norm.'
)
# PPO hps.
_GAE_LAMBDA = flags.DEFINE_float(
    'gae_lambda', 0.95, 'General advantage estimation lambda.'
)
_CLIPPING_EPSILON = flags.DEFINE_float(
    'clipping_epsilon', 0.3, 'Policy loss clipping epsilon.'
)
_NUM_RESETS_PER_EVAL = flags.DEFINE_integer(
    'num_resets_per_eval', 10, 'Number of resets per eval.'
)
_PPO_POLICY_HIDDEN_LAYER_SIZES = flags.DEFINE_string(
    'ppo_policy_hidden_layer_sizes', None, 'PPO policy hidden layer sizes.'
)
_PPO_VALUE_HIDDEN_LAYER_SIZES = flags.DEFINE_string(
    'ppo_value_hidden_layer_sizes', None, 'PPO value hidden layer sizes.'
)
_PPO_POLICY_OBS_KEY = flags.DEFINE_string(
    'ppo_policy_obs_key', None, 'PPO policy obs key.'
)
_PPO_VALUE_OBS_KEY = flags.DEFINE_string(
    'ppo_value_obs_key', None, 'PPO value obs key.'
)
# ARS hps.
_NUMBER_OF_DIRECTIONS = flags.DEFINE_integer(
    'number_of_directions',
    60,
    'Number of directions to explore. The actual number is 2x '
    'larger (used for antithetic sampling.',
)
_TOP_DIRECTIONS = flags.DEFINE_integer(
    'top_directions', 20, 'Number of top directions to select.'
)
_EXPLORATION_NOISE_STD = flags.DEFINE_float(
    'exploration_noise_std', 0.1, 'Std of a random noise added by ARS.'
)
_REWARD_SHIFT = flags.DEFINE_float(
    'reward_shift', 0.0, 'A reward shift to get rid of "stay alive" bonus.'
)
# ARS hps.
_POLICY_UPDATES = flags.DEFINE_integer(
    'policy_updates', None, 'Number of policy updates in APG.'
)
# Wrap.
_CUSTOM_WRAP_ENV = flags.DEFINE_bool(
    'custom_wrap_env', False, 'Wrap the environment with a custom wrapper.'
)


def get_env_factory(env_name: str):
  """Returns a function that creates an environment."""
  wrap_fn = None
  randomizer_fn = None
  if _CUSTOM_WRAP_ENV.value:
    pass
  else:
    get_environment = functools.partial(
        envs.get_environment, backend=_BACKEND.value
    )
  return get_environment, wrap_fn, randomizer_fn


def main(unused_argv):
  logdir = _LOGDIR.value

  ckpt_dir = epath.Path(logdir) / 'checkpoints'
  ckpt_dir.mkdir(exist_ok=True)
  get_environment, wrap_fn, randomizer_fn = get_env_factory(_ENV.value)
  with metrics.Writer(logdir) as writer:
    writer.write_hparams({
        'num_evals': _NUM_EVALS.value,
        'num_envs': _NUM_ENVS.value,
        'total_env_steps': _TOTAL_ENV_STEPS.value,
    })
    if _LEARNER.value == 'sac':
      network_factory = sac_networks.make_sac_networks
      if _Q_NETWORK_LAYER_NORM.value:
        network_factory = functools.partial(
            sac_networks.make_sac_networks, q_network_layer_norm=True
        )
      make_policy, params, _ = sac.train(
          environment=get_environment(_ENV.value),
          eval_env=get_environment(_ENV.value),
          wrap_env_fn=wrap_fn,
          randomization_fn=randomizer_fn,
          num_envs=_NUM_ENVS.value,
          action_repeat=_ACTION_REPEAT.value,
          normalize_observations=_NORMALIZE_OBSERVATIONS.value,
          num_timesteps=_TOTAL_ENV_STEPS.value,
          num_evals=_NUM_EVALS.value,
          batch_size=_BATCH_SIZE.value,
          min_replay_size=_MIN_REPLAY_SIZE.value,
          max_replay_size=_MAX_REPLAY_SIZE.value,
          network_factory=network_factory,
          learning_rate=_LEARNING_RATE.value,
          discounting=_DISCOUNTING.value,
          max_devices_per_host=_MAX_DEVICES_PER_HOST.value,
          seed=_SEED.value,
          reward_scaling=_REWARD_SCALING.value,
          grad_updates_per_step=_GRAD_UPDATES_PER_STEP.value,
          episode_length=_EPISODE_LENGTH.value,
          progress_fn=writer.write_scalars,
      )
    elif _LEARNER.value == 'es':
      make_policy, params, _ = es.train(
          environment=get_environment(_ENV.value),
          eval_env=get_environment(_ENV.value),
          wrap_env_fn=wrap_fn,
          randomization_fn=randomizer_fn,
          num_timesteps=_TOTAL_ENV_STEPS.value,
          fitness_shaping=es.FitnessShaping[_FITNESS_SHAPING.value.upper()],
          population_size=_POPULATION_SIZE.value,
          perturbation_std=_PERTURBATION_STD.value,
          normalize_observations=_NORMALIZE_OBSERVATIONS.value,
          action_repeat=_ACTION_REPEAT.value,
          num_evals=_NUM_EVALS.value,
          center_fitness=_CENTER_FITNESS.value,
          l2coeff=_L2COEFF.value,
          learning_rate=_LEARNING_RATE.value,
          seed=_SEED.value,
          max_devices_per_host=_MAX_DEVICES_PER_HOST.value,
          episode_length=_EPISODE_LENGTH.value,
          progress_fn=writer.write_scalars,
      )
    elif _LEARNER.value == 'ppo':
      network_factory = ppo_networks.make_ppo_networks
      if _PPO_POLICY_HIDDEN_LAYER_SIZES.value is not None:
        policy_hidden_layer_sizes = [
            int(x) for x in _PPO_POLICY_HIDDEN_LAYER_SIZES.value.split(',')
        ]
        network_factory = functools.partial(
            network_factory,
            policy_hidden_layer_sizes=policy_hidden_layer_sizes,
        )
      if _PPO_VALUE_HIDDEN_LAYER_SIZES.value is not None:
        value_hidden_layer_sizes = [
            int(x) for x in _PPO_VALUE_HIDDEN_LAYER_SIZES.value.split(',')
        ]
        network_factory = functools.partial(
            network_factory,
            value_hidden_layer_sizes=value_hidden_layer_sizes,
        )
      if _PPO_POLICY_OBS_KEY.value is not None:
        network_factory = functools.partial(
            network_factory,
            policy_obs_key=_PPO_POLICY_OBS_KEY.value,
        )
      if _PPO_VALUE_OBS_KEY.value is not None:
        network_factory = functools.partial(
            network_factory,
            value_obs_key=_PPO_VALUE_OBS_KEY.value,
        )
      make_policy, params, _ = ppo.train(
          environment=get_environment(_ENV.value),
          eval_env=get_environment(_ENV.value),
          wrap_env_fn=wrap_fn,
          randomization_fn=randomizer_fn,
          num_timesteps=_TOTAL_ENV_STEPS.value,
          episode_length=_EPISODE_LENGTH.value,
          network_factory=network_factory,
          action_repeat=_ACTION_REPEAT.value,
          num_envs=_NUM_ENVS.value,
          max_devices_per_host=_MAX_DEVICES_PER_HOST.value,
          learning_rate=_LEARNING_RATE.value,
          entropy_cost=_ENTROPY_COST.value,
          discounting=_DISCOUNTING.value,
          seed=_SEED.value,
          unroll_length=_UNROLL_LENGTH.value,
          batch_size=_BATCH_SIZE.value,
          num_minibatches=_NUM_MINIBATCHES.value,
          normalize_observations=_NORMALIZE_OBSERVATIONS.value,
          num_updates_per_batch=_NUM_UPDATES_PER_BATCH.value,
          num_evals=_NUM_EVALS.value,
          reward_scaling=_REWARD_SCALING.value,
          gae_lambda=_GAE_LAMBDA.value,
          clipping_epsilon=_CLIPPING_EPSILON.value,
          num_resets_per_eval=_NUM_RESETS_PER_EVAL.value,
          progress_fn=writer.write_scalars,
          save_checkpoint_path=ckpt_dir.as_posix(),
      )
    elif _LEARNER.value == 'apg':
      make_policy, params, _ = apg.train(
          environment=get_environment(_ENV.value),
          eval_env=get_environment(_ENV.value),
          wrap_env_fn=wrap_fn,
          randomization_fn=randomizer_fn,
          policy_updates=_POLICY_UPDATES.value,
          num_envs=_NUM_ENVS.value,
          action_repeat=_ACTION_REPEAT.value,
          num_evals=_NUM_EVALS.value,
          learning_rate=_LEARNING_RATE.value,
          seed=_SEED.value,
          max_devices_per_host=_MAX_DEVICES_PER_HOST.value,
          normalize_observations=_NORMALIZE_OBSERVATIONS.value,
          max_gradient_norm=_MAX_GRADIENT_NORM.value,
          episode_length=_EPISODE_LENGTH.value,
          progress_fn=writer.write_scalars,
      )
    elif _LEARNER.value == 'ars':
      make_policy, params, _ = ars.train(
          environment=get_environment(_ENV.value),
          eval_env=get_environment(_ENV.value),
          wrap_env_fn=wrap_fn,
          randomization_fn=randomizer_fn,
          number_of_directions=_NUMBER_OF_DIRECTIONS.value,
          max_devices_per_host=_MAX_DEVICES_PER_HOST.value,
          action_repeat=_ACTION_REPEAT.value,
          normalize_observations=_NORMALIZE_OBSERVATIONS.value,
          num_timesteps=_TOTAL_ENV_STEPS.value,
          exploration_noise_std=_EXPLORATION_NOISE_STD.value,
          num_evals=_NUM_EVALS.value,
          seed=_SEED.value,
          step_size=_LEARNING_RATE.value,
          top_directions=_TOP_DIRECTIONS.value,
          reward_shift=_REWARD_SHIFT.value,
          episode_length=_EPISODE_LENGTH.value,
          progress_fn=writer.write_scalars,
      )
    else:
      raise ValueError(f'Unknown learner: {_LEARNER.value}')

  # Save to flax serialized checkpoint.
  filename = f'{_ENV.value}_{_LEARNER.value}.pkl'
  path = os.path.join(logdir, filename)
  model.save_params(path, params)

  # Output an episode trajectory.
  get_environment, *_ = get_env_factory(_ENV.value)
  env = get_environment(_ENV.value)

  def do_rollout(rng, state):
    data_attr_name = 'pipeline_state' if hasattr(env, 'sys') else 'data'
    empty_data = getattr(state, data_attr_name).__class__(
        **{k: None for k in getattr(state, data_attr_name).__annotations__}
    )  # pytype: disable=attribute-error
    empty_traj = state.__class__(**{k: None for k in state.__annotations__})  # pytype: disable=attribute-error
    empty_traj = empty_traj.replace(**{data_attr_name: empty_data})

    def step(carry, _):
      state, rng = carry
      rng, act_key = jax.random.split(rng)
      act = make_policy(params)(state.obs, act_key)[0]
      state = env.step(state, act)
      if hasattr(state, 'data'):
        # select a sub-set of the data for playground envs
        traj_data = empty_traj.tree_replace({
            'data.qpos': state.data.qpos,
            'data.qvel': state.data.qvel,
            'data.time': state.data.time,
            'data.ctrl': state.data.ctrl,
            'data.mocap_pos': state.data.mocap_pos,
            'data.mocap_quat': state.data.mocap_quat,
            'data.xfrc_applied': state.data.xfrc_applied,
        })
      elif hasattr(state, 'pipeline_state'):
        # select the entire state for brax envs
        traj_data = empty_traj.replace(
            **{data_attr_name: getattr(state, data_attr_name)}
        )
      else:
        raise ValueError(
            f'Unknown data attribute name: {data_attr_name} on state: {state}.'
        )
      return (state, rng), traj_data

    _, traj = jax.lax.scan(
        step, (state, rng), None, length=_EPISODE_LENGTH.value
    )
    return traj

  rng = jax.random.split(jax.random.PRNGKey(_SEED.value), _NUM_VIDEOS.value)
  reset_states = jax.jit(jax.vmap(env.reset))(rng)
  traj_stacked = jax.jit(jax.vmap(do_rollout))(rng, reset_states)
  trajectories = [None] * _NUM_VIDEOS.value
  for i in range(_NUM_VIDEOS.value):
    t = jax.tree.map(lambda x, i=i: x[i], traj_stacked)
    trajectories[i] = [
        jax.tree.map(lambda x, j=j: x[j], t)
        for j in range(_EPISODE_LENGTH.value)
    ]

  video_path = ''
  if hasattr(env, 'render'):
    for i in range(_NUM_VIDEOS.value):
      path_ = epath.Path(f'{logdir}/saved_videos/trajectory_{i:04d}.mp4')
      path_.parent.mkdir(parents=True, exist_ok=True)
      frames = env.render(trajectories[i])
      media.write_video(path_, frames, fps=1.0 / env.dt)
      video_path = path_.as_posix()
  elif _NUM_VIDEOS.value > 0:
    logging.warn('Cannot save videos for non physics environments.')



if __name__ == '__main__':
  app.run(main)



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Losses for BC."""

from typing import Any, Callable, Dict, Tuple

from brax.training.agents.bc import networks
from brax.training.types import Params
import jax.numpy as jp


# Vanilla L2 with postprocessing
def bc_loss(
    params: Params,
    normalizer_params: Any,
    data: Dict,
    make_policy: Callable[[Tuple[Any, Params]], networks.BCInferenceFn],
):
  policy = make_policy((normalizer_params, params))
  _, action_extras = policy(data['observations'], key_sample=None)  # pytype: disable=wrong-keyword-args
  actor_loss = (
      (
          (
              jp.tanh(action_extras['loc'])
              - jp.tanh(data['teacher_action_extras']['loc'])
          )
          ** 2
      )
      .sum(-1)
      .mean()
  )
  actor_loss = actor_loss.mean()
  return actor_loss, {'actor_loss': actor_loss, 'mse_loss': actor_loss}



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BC training."""

import functools
import time
from typing import Any, Callable, Optional, Tuple

from absl import logging
from brax import envs
from brax.training import acting
from brax.training import gradients
from brax.training import types
from brax.training.acme import running_statistics
from brax.training.acme import specs
from brax.training.agents.bc import checkpoint
from brax.training.agents.bc import losses as bc_losses
from brax.training.agents.bc import networks as bc_networks
from brax.training.agents.ppo import train as ppo_train
from brax.training.types import Params
from brax.training.types import PRNGKey
import flax
import jax
import jax.numpy as jp
import optax


@jax.vmap
def _random_crop(key, img):
  """Expects H W C image.

  Adapted from
  https://github.com/ikostrikov/jaxrl/blob/main/jaxrl/agents/drq/augmentations.py

  Returns:
    cropped image with shape [H, W, C]
  """
  padding = 2 if img.shape[-2] == 32 else 4
  crop_from = jax.random.randint(key, (2,), 0, 2 * padding + 1)
  crop_from = jp.concatenate([crop_from, jp.zeros((1,), dtype=jp.int32)])
  padded_img = jp.pad(
      img, ((padding, padding), (padding, padding), (0, 0)), mode='edge'
  )
  return jax.lax.dynamic_slice(padded_img, crop_from, img.shape)


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""

  # Optimisations:
  # 1) No scanning over trajectory collection. Just have a lot of parallel envs.
  # 2) Data never leaves the GPU. Only parameters flow in and out of it.
  optimizer_state: optax.OptState
  params: Params
  normalizer_params: running_statistics.RunningStatisticsState
  dagger_step: jp.ndarray


Metrics = types.Metrics


def train(
    demo_length: int,
    teacher_inference_fn: Callable[
        [types.Observation, Optional[PRNGKey]], Tuple[jp.ndarray, types.Extra]
    ],
    normalize_observations: bool = True,
    epochs: int = 50,
    tanh_squash: bool = True,
    env: Optional[envs.Env] = None,
    num_envs: int = 0,
    num_eval_envs: int = 0,
    eval_length: Optional[int] = None,
    batch_size: int = 256,
    scramble_time: int = 0,
    network_factory: types.NetworkFactory[
        bc_networks.BCNetworks
    ] = bc_networks.make_bc_networks,
    progress_fn: Callable[[int, Metrics], None] = lambda *args: None,
    madrona_backend: bool = False,
    seed: int = 0,
    learning_rate=4e-4,
    dagger_steps: int = 1,
    dagger_beta_fn: Callable[[int], float] = lambda step: jp.where(
        step == 0, 1.0, 0.0
    ),
    num_evals=0,
    augment_pixels: bool = False,
    reset: bool = True,
    save_checkpoint_path: Optional[str] = None,
    restore_checkpoint_path: Optional[str] = None,
    restore_params: Optional[Any] = None,
):
  """Online DAgger behavior cloning training.

  Args:
    demo_length: the length of demonstration trajectories
    teacher_inference_fn: function that generates teacher actions
    normalize_observations: whether to normalize observations
    epochs: number of supervised training epochs per DAgger iteration
    tanh_squash: whether to apply tanh squashing to actions. Improves training
      stability.
    env: the environment to train in
    num_envs: the number of parallel environments to use for rollouts
    num_eval_envs: the number of envs to use for evaluation
    eval_length: the length of an evaluation episode
    batch_size: the batch size for each training step
    scramble_time: Maximum time to scramble the envs by. Staggering initial
      times encourages a stationary distribution. This smoothes loss curves.
    network_factory: function that generates networks for policy
    progress_fn: a user-defined callback function for reporting/plotting metrics
    madrona_backend: whether to use Madrona backend for training
    seed: random seed
    learning_rate: learning rate for optimizer
    dagger_steps: number of DAgger iterations to perform
    dagger_beta_fn: function that determines probability of using teacher
      actions
    num_evals: the number of evals to run during the entire training run.
      Increasing the number of evals increases total training time
    augment_pixels: whether to add image augmentation to pixel inputs
    reset: whether to periodically use true resets for additional randomness
    save_checkpoint_path: the path used to save checkpoints. If None, no
      checkpoints are saved.
    restore_checkpoint_path: the path used to restore previous model params
    restore_params: raw network parameters to restore the TrainingState from.
      These override `restore_checkpoint_path`. These paramaters can be obtained
      from the return values of bc.train().

  Assumes your env is already wrapped.

  Returns:
    A tuple of (make_policy, training_state, metrics).
  """
  if env is None:
    raise ValueError('env must be set')

  if madrona_backend:
    if num_eval_envs and num_eval_envs != num_envs:
      raise ValueError('Madrona-MJX requires a fixed batch size')
    else:
      num_eval_envs = num_envs

  if num_evals > 0:
    assert eval_length is not None, 'eval_length must be set if num_evals > 0'
  num_evals_after_init = max(num_evals - 1, 1)
  assert (
      dagger_steps % num_evals_after_init == 0
  ), 'Dagger steps must be divisible by num_evals - 1'
  dagger_steps_per_eval = dagger_steps / num_evals_after_init

  key = jax.random.PRNGKey(seed)

  def scramble_times(key, state):
    state.info['steps'] = jax.random.randint(
        key, (num_envs,), minval=0, maxval=scramble_time
    )
    if '_steps' in state.info:  # Some envs have internal clocks.
      state.info['_steps'] = jp.array(state.info['steps'])
    return state

  # Init the env state early to use for obs size calculations.
  key, key_reset, key_scramble = jax.random.split(key, 3)
  reset_fn = jax.jit(env.reset)
  env_state = reset_fn(jax.random.split(key_reset, num_envs))
  env_state = scramble_times(key_scramble, env_state)
  obs_shape = jax.tree_util.tree_map(
      lambda x: x.shape[1:], env_state.obs
  )  # Discard batch axis over envs.

  normalize = lambda x, y: x
  if normalize_observations:
    normalize = running_statistics.normalize
  bc_network = network_factory(
      obs_shape, env.action_size, preprocess_observations_fn=normalize
  )
  make_policy = functools.partial(
      bc_networks.make_inference_fn(bc_network),
      deterministic=True,
      tanh_squash=tanh_squash,
  )
  key_policy, key = jax.random.split(key)
  optimizer = optax.adam(learning_rate=learning_rate)
  specs_obs_shape = jax.tree_util.tree_map(
      lambda x: specs.Array(x.shape[-1:], jp.dtype('float32')), env_state.obs
  )
  init_params = bc_network.policy_network.init(key_policy)
  training_state = TrainingState(
      params=init_params,
      normalizer_params=running_statistics.init_state(
          ppo_train._remove_pixels(specs_obs_shape)
      ),
      optimizer_state=optimizer.init(
          init_params
      ),  # pytype: disable=wrong-arg-types  # numpy-scalars
      dagger_step=jp.array(0, dtype=int),
  )

  if restore_checkpoint_path is not None:
    params = checkpoint.load(restore_checkpoint_path)
    training_state = training_state.replace(
        normalizer_params=params[0],
        params=params[1],
    )

  if restore_params is not None:
    logging.info('Restoring TrainingState from `restore_params`.')
    training_state = training_state.replace(
        normalizer_params=restore_params[0],
        params=restore_params[1],
    )

  loss_fn = functools.partial(bc_losses.bc_loss, make_policy=make_policy)

  gradient_update_fn = jax.jit(
      gradients.gradient_update_fn(
          loss_fn, optimizer, pmap_axis_name=None, has_aux=True
      )
  )

  ckpt_config = checkpoint.network_config(
      observation_size=obs_shape,
      action_size=env.action_size,
      normalize_observations=normalize_observations,
      network_factory=network_factory,
  )

  def collect_unroll(
      env_state: envs.State,
      key: PRNGKey,
      params: Tuple[Params, Params],
      beta: float,
  ):
    """Collect demo trajectory of fixed length."""

    def env_step(carry: Tuple[PRNGKey, envs.State], _):
      key, env_state = carry
      key, key_turn = jax.random.split(key)
      student_inference_fn = make_policy(params)
      teachers_turn = jax.random.bernoulli(key_turn, beta, shape=(num_envs, 1))
      actions = jp.where(
          teachers_turn,
          teacher_inference_fn(env_state.obs, None)[0],  # E x action_size
          student_inference_fn(env_state.obs, None)[0],  # pytype: disable=wrong-arg-types
      )
      nstate = env.step(env_state, actions)
      return (key, nstate), (env_state.obs, env_state.reward)  # E x ...

    key_step, key = jax.random.split(key)
    (_, env_state), (obs, reward) = jax.lax.scan(
        env_step, (key_step, env_state), length=demo_length
    )  # T x E x ...
    obs = jax.tree_util.tree_map(lambda x: jp.swapaxes(x, 0, 1), obs)
    reward = jp.swapaxes(reward, 0, 1)
    assert list(obs.values())[0].shape[:2] == (num_envs, demo_length), obs.shape
    return env_state, obs, reward  # E x T

  def collect_data(
      env_state: envs.State,
      key: PRNGKey,
      params: Tuple[Params, Params],
      beta: float,
  ):
    """Collect a dataset for behavioural cloning.

    beta = probability of taking teacher action.
    """
    key, key_unroll = jax.random.split(key)
    env_state, all_obs, all_rewards = collect_unroll(
        env_state, key_unroll, params, beta
    )  # Tree(E x T x ...)

    for k in all_obs.keys():
      assert all_obs[k].shape[:2] == (num_envs, demo_length), (
          f'Expected shape {(num_envs, demo_length)} but got'
          f' {all_obs[k].shape[:2]}'
      )

    metrics = {
        'reward_mean': jp.mean(all_rewards, axis=0),
        'reward_std': jp.std(all_rewards, axis=0),
    }
    return env_state, all_obs, metrics

  def combine_batch_axes(tree):
    return jax.tree_util.tree_map(
        lambda x: x.reshape((-1,) + x.shape[2:]), tree
    )

  assert (
      int(num_envs * demo_length) % batch_size == 0
  ), 'Demo trajectories must be evenly divisible into batches'
  num_minibatches = int(num_envs * demo_length / batch_size)

  def fit_student(ts: TrainingState, X, key) -> Tuple[TrainingState, Metrics]:
    # Fixed through fitting process.
    normalizer_params = ts.normalizer_params

    def epoch(carry, _):
      params, optimizer_state, key = carry
      key, key_perm = jax.random.split(key)
      key, key_crop = jax.random.split(key)
      # Correlated crop between views for multi-pixel case.
      key_crop = jax.random.split(key_crop, int(num_envs * demo_length))

      shifted = X
      if augment_pixels:
        shifted = {}
        for k, v in X.items():
          if k.startswith('pixels/view'):
            shifted[k] = _random_crop(key_crop, v)
        shifted = {**X, **shifted}  # shared keys are overwritten by second arg.

      def convert_data(x: jp.ndarray):
        x = jax.random.permutation(key_perm, x)
        x = jp.reshape(x, (num_minibatches, -1) + x.shape[1:])
        return x

      X_shuffled = jax.tree_util.tree_map(
          convert_data, shifted
      )  # num_minibatches, minibatch_size, ...

      def minibatch_step(carry, X_batch):
        params, optimizer_state = carry
        t_act, t_act_extras = teacher_inference_fn(X_batch, None)
        data = {
            'observations': X_batch,
            'teacher_action': t_act,
            'teacher_action_extras': t_act_extras,
        }
        # Apply loss grad fn
        (_, metrics), params, optimizer_state = gradient_update_fn(
            params,
            normalizer_params,
            data,
            optimizer_state=optimizer_state,
        )
        return (params, optimizer_state), metrics

      (params, optimizer_state), metrics = jax.lax.scan(
          minibatch_step, (params, optimizer_state), X_shuffled
      )
      return (params, optimizer_state, key), metrics

    key, key_epoch = jax.random.split(key)
    (params, optimizer_state, _), metrics = jax.lax.scan(
        epoch, (ts.params, ts.optimizer_state, key_epoch), length=epochs
    )
    return ts.replace(optimizer_state=optimizer_state, params=params), metrics

  @jax.jit
  def dagger_step(
      carry: Tuple[PRNGKey, envs.State, TrainingState], _
  ) -> Tuple[Tuple[PRNGKey, envs.State, TrainingState], Metrics]:
    # 1 dagger epoch.
    key, env_state, ts = carry
    key, key_data = jax.random.split(key)
    env_state, raw_data, data_metrics = collect_data(
        env_state,
        key_data,
        (ts.normalizer_params, ts.params),
        dagger_beta_fn(ts.dagger_step),  # pytype: disable=wrong-arg-types
    )
    assert data_metrics['reward_mean'].shape == (demo_length,), (
        f'Expected shape {(demo_length,)} but got'
        f' {data_metrics["reward_mean"].shape}'
    )

    X_cur = combine_batch_axes(raw_data)
    ts = ts.replace(
        normalizer_params=running_statistics.update(
            ts.normalizer_params,
            ppo_train._remove_pixels(X_cur),
            pmap_axis_name=None,
        )
    )
    key, key_fit = jax.random.split(key)
    ts, metrics = fit_student(ts, X_cur, key_fit)
    metrics = {**data_metrics, **metrics}
    return (key, env_state, ts.replace(dagger_step=ts.dagger_step + 1)), metrics

  def evaluate(training_state: TrainingState, other_metrics: Metrics):
    inference_params = (training_state.normalizer_params, training_state.params)
    if num_evals:
      eval_metrics = evaluator.run_evaluation(
          inference_params,
          training_metrics={},
      )
      other_metrics.update(eval_metrics)  # pytype: disable=attribute-error

    if save_checkpoint_path is not None:
      checkpoint.save(  # pytype: disable=wrong-arg-types
          save_checkpoint_path,
          training_state.dagger_step,
          inference_params,
          ckpt_config,
      )

    progress_fn(training_state.dagger_step, other_metrics)  # pytype: disable=wrong-arg-types

  if num_evals:
    eval_env = env
    Evaluator = acting.Evaluator
    key, eval_key = jax.random.split(key)
    evaluator = Evaluator(
        eval_env,
        make_policy,
        num_eval_envs=num_eval_envs,
        episode_length=eval_length,
        action_repeat=1,
        key=eval_key,
    )
    evaluate(training_state, {})

  wall_time = 0
  key, key_dagger = jax.random.split(key)
  for i_eval in range(num_evals_after_init):
    if i_eval != 0 and reset:
      # Seed new randomness to be gradually phased in.
      key_reset, key = jax.random.split(key)
      env_state = reset_fn(jax.random.split(key_reset, num_envs))

    t0 = time.monotonic()
    (key_dagger, env_state, training_state), b_metrics = jax.lax.scan(
        dagger_step,
        (key_dagger, env_state, training_state),
        length=dagger_steps_per_eval,
    )
    jax.tree_util.tree_map(lambda x: x.block_until_ready(), b_metrics)
    t1 = time.monotonic()
    d_length = int(num_envs * demo_length)
    b_metrics['SPS'] = d_length * epochs / (t1 - t0)
    wall_time += t1 - t0
    b_metrics['walltime'] = wall_time
    evaluate(training_state, b_metrics)

  return (
      make_policy,
      (training_state.normalizer_params, training_state.params),
      b_metrics,
  )



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PPO tests."""

import functools
import pickle
from absl.testing import absltest
from absl.testing import parameterized
from brax import envs
from brax.training.acme import running_statistics
from brax.training.agents.ppo import networks as ppo_networks
from brax.training.agents.ppo import networks_vision as ppo_networks_vision
from brax.training.agents.ppo import train as ppo
import jax


class PPOTest(parameterized.TestCase):
  """Tests for PPO module."""


  @parameterized.parameters('ndarray', 'dict_state')
  def testTrain(self, obs_mode):
    """Test PPO with a simple env."""
    fast = envs.get_environment('fast', obs_mode=obs_mode)
    _, _, metrics = ppo.train(
        fast,
        num_timesteps=2**15,
        episode_length=128,
        num_envs=64,
        learning_rate=3e-4,
        entropy_cost=1e-2,
        discounting=0.95,
        unroll_length=5,
        batch_size=64,
        num_minibatches=8,
        num_updates_per_batch=4,
        normalize_observations=True,
        seed=2,
        num_evals=3,
        reward_scaling=10,
        normalize_advantage=False,
    )
    self.assertGreater(metrics['eval/episode_reward'], 135)
    self.assertEqual(fast.reset_count, 2)  # type: ignore
    self.assertEqual(fast.step_count, 2)  # type: ignore

  @parameterized.parameters(
      ('normal', 'scalar'),
      ('normal', 'log'),
      ('tanh_normal', 'log'),
  )
  def testTrainWithNetworkParams(self, distribution_type, noise_std_type):
    """Test PPO runs with different network params."""
    network_factory = functools.partial(
        ppo_networks.make_ppo_networks,
        distribution_type=distribution_type,
        noise_std_type=noise_std_type,
    )

    _, _, _ = ppo.train(
        envs.get_environment('inverted_pendulum', backend='spring'),
        num_timesteps=2**13,
        episode_length=50,
        num_envs=64,
        learning_rate=3e-4,
        entropy_cost=1e-2,
        discounting=0.95,
        unroll_length=5,
        batch_size=64,
        num_minibatches=8,
        num_updates_per_batch=4,
        normalize_observations=True,
        seed=2,
        reward_scaling=10,
        normalize_advantage=False,
        network_factory=network_factory,
    )

  def testTrainAsymmetricActorCritic(self):
    """Test PPO with asymmetric actor critic."""
    env = envs.get_environment(
        'fast', asymmetric_obs=True, obs_mode='dict_state'
    )

    network_factory = functools.partial(
        ppo_networks.make_ppo_networks,
        policy_hidden_layer_sizes=(32,),
        value_hidden_layer_sizes=(32,),
        policy_obs_key='state',
        value_obs_key='privileged_state',
    )

    _, (_, policy_params, value_params), _ = ppo.train(
        env,
        num_timesteps=2**15,
        episode_length=1000,
        num_envs=64,
        learning_rate=3e-4,
        entropy_cost=1e-2,
        discounting=0.95,
        unroll_length=5,
        batch_size=64,
        num_minibatches=8,
        num_updates_per_batch=4,
        normalize_observations=False,
        seed=2,
        reward_scaling=10,
        normalize_advantage=False,
        network_factory=network_factory,
    )

    self.assertEqual(
        policy_params['params']['hidden_0']['kernel'].shape,
        (env.observation_size['state'], 32),
    )
    self.assertEqual(
        value_params['params']['hidden_0']['kernel'].shape,
        (env.observation_size['privileged_state'], 32),
    )

  @parameterized.parameters(True, False)
  def testNetworkEncoding(self, normalize_observations):
    env = envs.get_environment('fast')
    original_inference, params, _ = ppo.train(
        env,
        num_timesteps=128,
        episode_length=128,
        num_envs=128,
        normalize_observations=normalize_observations,
    )
    normalize_fn = lambda x, y: x
    if normalize_observations:
      normalize_fn = running_statistics.normalize
    ppo_network = ppo_networks.make_ppo_networks(
        env.observation_size, env.action_size, normalize_fn
    )
    inference = ppo_networks.make_inference_fn(ppo_network)
    byte_encoding = pickle.dumps(params)
    decoded_params = pickle.loads(byte_encoding)

    # Compute one action.
    state = env.reset(jax.random.PRNGKey(0))
    original_action = original_inference(decoded_params)(
        state.obs, jax.random.PRNGKey(0)
    )[0]
    action = inference(decoded_params)(state.obs, jax.random.PRNGKey(0))[0]
    self.assertSequenceEqual(original_action, action)
    env.step(state, action)

  def testTrainDomainRandomize(self):
    """Test PPO with domain randomization."""

    def rand_fn(sys, rng):
      @jax.vmap
      def get_offset(rng):
        offset = jax.random.uniform(rng, shape=(3,), minval=-0.1, maxval=0.1)
        pos = sys.link.transform.pos.at[0].set(offset)
        return pos

      sys_v = sys.tree_replace({'link.inertia.transform.pos': get_offset(rng)})
      in_axes = jax.tree.map(lambda x: None, sys)
      in_axes = in_axes.tree_replace({'link.inertia.transform.pos': 0})
      return sys_v, in_axes

    _, _, _ = ppo.train(
        envs.get_environment('inverted_pendulum', backend='spring'),
        num_timesteps=2**15,
        episode_length=1000,
        num_envs=64,
        learning_rate=3e-4,
        entropy_cost=1e-2,
        discounting=0.95,
        unroll_length=5,
        batch_size=64,
        num_minibatches=8,
        num_updates_per_batch=4,
        normalize_observations=True,
        seed=2,
        reward_scaling=10,
        normalize_advantage=False,
        randomization_fn=rand_fn,
    )

  @parameterized.parameters(
      {'asymmetric_obs': True, 'obs_mode': 'dict_pixels_state'},
      {'asymmetric_obs': False, 'obs_mode': 'dict_pixels_state'},
      {'asymmetric_obs': False, 'obs_mode': 'dict_pixels'},
  )
  def testPixelsPPO(self, asymmetric_obs, obs_mode):
    """Test PPO with pixel observations."""
    env = envs.get_environment(
        'fast',
        pixel_obs=True,
        asymmetric_obs=asymmetric_obs,
        obs_mode=obs_mode,
    )
    if obs_mode == 'dict_pixels':
      policy_obs_key = ''
      value_obs_key = ''
    else:
      policy_obs_key = 'state'
      value_obs_key = 'privileged_state' if asymmetric_obs else 'state'

    network_factory = functools.partial(
        ppo_networks_vision.make_ppo_networks_vision,
        policy_hidden_layer_sizes=(32,),
        value_hidden_layer_sizes=(32,),
        policy_obs_key=policy_obs_key,
        value_obs_key=value_obs_key,
    )

    _, (_, policy_params, value_params), _ = ppo.train(
        env,
        num_timesteps=2**15,
        episode_length=1000,
        num_envs=64,
        learning_rate=3e-4,
        entropy_cost=1e-2,
        discounting=0.95,
        unroll_length=5,
        batch_size=64,
        num_minibatches=8,
        num_updates_per_batch=4,
        normalize_observations=True,
        seed=2,
        reward_scaling=10,
        normalize_advantage=False,
        network_factory=network_factory,
        augment_pixels=True,
    )
    num_views = 2
    cnn_features = 64

    if asymmetric_obs:
      self.assertEqual(
          policy_params['params']['MLP_0']['hidden_0']['kernel'].shape,
          (num_views * cnn_features + env.observation_size['state'], 32),
      )
      self.assertEqual(
          value_params['params']['MLP_0']['hidden_0']['kernel'].shape,
          (
              num_views * cnn_features
              + env.observation_size['privileged_state'],
              32,
          ),
      )
    if obs_mode == 'dict_pixels':
      self.assertEqual(
          policy_params['params']['MLP_0']['hidden_0']['kernel'].shape,
          (num_views * cnn_features, 32),
      )


if __name__ == '__main__':
  jax.config.update('jax_threefry_partitionable', False)
  absltest.main()



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Checkpointing for PPO."""

from typing import Any, Union

from brax.training import checkpoint
from brax.training import types
from brax.training.agents.ppo import networks as ppo_networks
from etils import epath
from ml_collections import config_dict

_CONFIG_FNAME = 'ppo_network_config.json'


def save(
    path: Union[str, epath.Path],
    step: int,
    params: Any,
    config: config_dict.ConfigDict,
):
  """Saves a checkpoint."""
  return checkpoint.save(path, step, params, config, _CONFIG_FNAME)


def load(
    path: Union[str, epath.Path],
):
  """Loads checkpoint."""
  return checkpoint.load(path)


def network_config(
    observation_size: types.ObservationSize,
    action_size: int,
    normalize_observations: bool,
    network_factory: types.NetworkFactory[ppo_networks.PPONetworks],
) -> config_dict.ConfigDict:
  """Returns a config dict for re-creating a network from a checkpoint."""
  return checkpoint.network_config(
      observation_size, action_size, normalize_observations, network_factory
  )


def _get_ppo_network(
    config: config_dict.ConfigDict,
    network_factory: types.NetworkFactory[ppo_networks.PPONetworks],
) -> ppo_networks.PPONetworks:
  """Generates a PPO network given config."""
  return checkpoint.get_network(config, network_factory)  # pytype: disable=bad-return-type


def load_config(
    path: Union[str, epath.Path],
) -> config_dict.ConfigDict:
  """Loads PPO config from checkpoint."""
  path = epath.Path(path)
  config_path = path / _CONFIG_FNAME
  return checkpoint.load_config(config_path)


def load_policy(
    path: Union[str, epath.Path],
    network_factory: types.NetworkFactory[
        ppo_networks.PPONetworks
    ] = ppo_networks.make_ppo_networks,
    deterministic: bool = True,
):
  """Loads policy inference function from PPO checkpoint."""
  path = epath.Path(path)
  config = load_config(path)
  params = load(path)
  ppo_network = _get_ppo_network(config, network_factory)
  make_inference_fn = ppo_networks.make_inference_fn(ppo_network)

  return make_inference_fn(params, deterministic=deterministic)



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Gotta go fast!  This trivial Env is for unit testing."""

import enum

from brax import base
from brax.envs.base import PipelineEnv, State
import jax
from jax import numpy as jp


class ObservationMode(enum.Enum):
  """Describes observation formats.

  Attributes:
    NDARRAY: Flat NumPy array of state info.
    DICT_STATE: Dictionary of state info.
    DICT_PIXELS: Dictionary of pixel observations.
    DICT_PIXELS_STATE: Dictionary of both state and pixel info.
    DICT_LATENT_STATE: Dictionary with latent pixel and state info.
  """

  NDARRAY = 'ndarray'
  DICT_STATE = 'dict_state'
  DICT_PIXELS = 'dict_pixels'
  DICT_PIXELS_STATE = 'dict_pixels_state'
  DICT_LATENT_STATE = 'dict_latent_state'


class Fast(PipelineEnv):
  """Trains an agent to go fast."""

  def __init__(
      self,
      asymmetric_obs: bool = False,
      obs_mode: ObservationMode = ObservationMode.NDARRAY,
      **kwargs,
  ):
    self._dt = 0.02
    self._reset_count = 0
    self._step_count = 0
    self._asymmetric_obs = asymmetric_obs
    self._obs_mode = ObservationMode(obs_mode)

    if self._asymmetric_obs and self._obs_mode == ObservationMode.NDARRAY:
      raise ValueError('asymmetric_obs requires dictionary observations')

  def reset(self, rng: jax.Array) -> State:
    self._reset_count += 1
    pipeline_state = base.State(
        q=jp.zeros(1),
        qd=jp.zeros(1),
        x=base.Transform.create(pos=jp.zeros(3)),
        xd=base.Motion.create(vel=jp.zeros(3)),
        contact=None,
    )
    obs = {'state': jp.zeros(2)}
    if self._asymmetric_obs:
      obs['privileged_state'] = jp.zeros(4)  # Dummy privileged state.
    pixels = {
        'pixels/view_0': jp.zeros((4, 4, 3)),
        'pixels/view_1': jp.zeros((4, 4, 3)),
    }
    latent_pixels = {
        'latent_pixels/view_0': jp.zeros(12),
        'latent_pixels/view_1': jp.zeros(12),
    }

    if self._obs_mode == ObservationMode.DICT_PIXELS:
      obs = pixels
    elif self._obs_mode == ObservationMode.DICT_PIXELS_STATE:
      obs = {**obs, **pixels}
    elif self._obs_mode == ObservationMode.DICT_LATENT_STATE:
      obs = {**obs, **latent_pixels}
    elif self._obs_mode == ObservationMode.NDARRAY:
      obs = obs['state']

    reward, done = jp.array(0.0), jp.array(0.0)
    return State(pipeline_state, obs, reward, done)

  def step(self, state: State, action: jax.Array) -> State:
    assert state.pipeline_state is not None
    self._step_count += 1
    vel = state.pipeline_state.xd.vel + (action > 0) * self._dt
    pos = state.pipeline_state.x.pos + vel * self._dt

    qp = state.pipeline_state.replace(
        x=state.pipeline_state.x.replace(pos=pos),
        xd=state.pipeline_state.xd.replace(vel=vel),
    )
    obs = {'state': jp.array([pos[0], vel[0]])}
    if self._asymmetric_obs:
      obs['privileged_state'] = jp.zeros(4)  # Dummy privileged state.
    pixels = {
        'pixels/view_0': jp.zeros((4, 4, 3)),
        'pixels/view_1': jp.zeros((4, 4, 3)),
    }
    latent_pixels = {
        'latent_pixels/view_0': jp.zeros(12),
        'latent_pixels/view_1': jp.zeros(12),
    }

    if self._obs_mode == ObservationMode.DICT_PIXELS:
      obs = pixels
    elif self._obs_mode == ObservationMode.DICT_PIXELS_STATE:
      obs = {**obs, **pixels}
    elif self._obs_mode == ObservationMode.DICT_LATENT_STATE:
      obs = {**obs, **latent_pixels}
    elif self._obs_mode == ObservationMode.NDARRAY:
      obs = obs['state']

    reward = pos[0]
    return state.replace(pipeline_state=qp, obs=obs, reward=reward)

  @property
  def reset_count(self):
    return self._reset_count

  @property
  def step_count(self):
    return self._step_count

  @property
  def observation_size(self):
    ret = super().observation_size
    if self._obs_mode == ObservationMode.NDARRAY:
      return ret
    # Turn 1-D tuples to ints.
    return {
        key: value[0] if len(value) == 1 else value
        for key, value in ret.items()  # pytype: disable=attribute-error
    }

  @property
  def action_size(self):
    return 1



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logger for training metrics."""

import collections
import logging
from jax import numpy as jnp
import numpy as np


class EpisodeMetricsLogger:
  """Logs training metrics for each episode."""

  def __init__(
      self, buffer_size=100, steps_between_logging=1e5, progress_fn=None
  ):
    self._metrics_buffer = collections.defaultdict(
        lambda: collections.deque(maxlen=buffer_size)
    )
    self._buffer_size = buffer_size
    self._steps_between_logging = steps_between_logging
    self._num_steps = 0
    self._last_log_steps = 0
    self._log_count = 0
    self._progress_fn = progress_fn

  def update_episode_metrics(self, metrics, dones):
    self._num_steps += np.prod(dones.shape)
    if jnp.sum(dones) > 0:
      for name, metric in metrics.items():
        done_metrics = metric[dones.astype(bool)].flatten().tolist()
        self._metrics_buffer[name].extend(done_metrics)
    if self._num_steps - self._last_log_steps >= self._steps_between_logging:
      self.log_metrics()
      self._last_log_steps = self._num_steps

  def log_metrics(self, pad=35):
    """Log metrics to console."""
    self._log_count += 1
    log_string = (
        f"\n{'Steps':>{pad}} Env: {self._num_steps} Log: {self._log_count}\n"
    )
    mean_metrics = {}
    for metric_name in self._metrics_buffer:
      mean_metrics[metric_name] = np.mean(self._metrics_buffer[metric_name])
      log_string += (
          f"{f'Episode {metric_name}:':>{pad}}"
          f" {mean_metrics[metric_name]:.4f}\n"
      )
    logging.info(log_string)
    if self._progress_fn is not None:
      self._progress_fn(
          int(self._num_steps),
          {f"episode/{name}": value for name, value in mean_metrics.items()},
      )



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Checkpointing functions."""

import inspect
import json
import logging
from typing import Any, Dict, Tuple, Union

from brax.training import types
from brax.training.acme import running_statistics
from brax.training.agents.bc import networks as bc_networks
from brax.training.agents.ppo import networks as ppo_networks
from brax.training.agents.sac import networks as sac_networks
from etils import epath
from flax.training import orbax_utils
import jax
from ml_collections import config_dict
import numpy as np
from orbax import checkpoint as ocp


def _get_function_kwargs(func: Any) -> Dict[str, Any]:
  """Gets kwargs of a function."""
  return {
      p.name: p.default
      for p in inspect.signature(func).parameters.values()
      if p.default is not inspect.Parameter.empty
  }


def _get_function_defaults(func: Any) -> Dict[str, Any]:
  """Gets default kwargs of a function potentially wrapped in partials."""
  kwargs = _get_function_kwargs(func)
  if hasattr(func, 'func'):
    kwargs.update(_get_function_defaults(func.func))
  return kwargs


def network_config(
    observation_size: types.ObservationSize,
    action_size: int,
    normalize_observations: bool,
    network_factory: types.NetworkFactory[
        Union[
            bc_networks.BCNetworks,
            ppo_networks.PPONetworks,
            sac_networks.SACNetworks,
        ]
    ],
) -> config_dict.ConfigDict:
  """Returns a config dict for re-creating a network from a checkpoint."""
  config = config_dict.ConfigDict()
  kwargs = _get_function_kwargs(network_factory)
  defaults = _get_function_defaults(network_factory)

  if 'preprocess_observations_fn' in kwargs:
    if (
        kwargs['preprocess_observations_fn']
        != defaults['preprocess_observations_fn']
    ):
      raise ValueError(
          'checkpointing only supports identity_observation_preprocessor as the'
          ' preprocess_observations_fn'
      )
    del kwargs['preprocess_observations_fn']
  if 'activation' in kwargs:
    # TODO: Add other activations.
    if kwargs['activation'] != defaults['activation']:
      raise ValueError('checkpointing only supports default activation')
    del kwargs['activation']

  config.network_factory_kwargs = kwargs
  config.normalize_observations = normalize_observations
  config.observation_size = observation_size
  config.action_size = action_size
  return config


def get_network(
    config: config_dict.ConfigDict,
    network_factory: types.NetworkFactory[
        Union[
            bc_networks.BCNetworks,
            ppo_networks.PPONetworks,
            sac_networks.SACNetworks,
        ]
    ],
) -> Union[
    bc_networks.BCNetworks, ppo_networks.PPONetworks, sac_networks.SACNetworks
]:
  """Generates a network given config."""
  normalize = lambda x, y: x
  if config.normalize_observations:
    normalize = running_statistics.normalize
  network = network_factory(
      config.to_dict()['observation_size'],
      config.action_size,
      preprocess_observations_fn=normalize,
      **config.network_factory_kwargs,
  )
  return network


def save(
    path: Union[str, epath.Path],
    step: int,
    params: Tuple[Any, ...],
    config: config_dict.ConfigDict,
    config_fname: str = 'config.json',
):
  """Saves a checkpoint."""
  ckpt_path = epath.Path(path) / f'{step:012d}'
  logging.info('saving checkpoint to %s', ckpt_path.as_posix())

  if not ckpt_path.exists():
    ckpt_path.mkdir(parents=True)

  orbax_checkpointer = ocp.PyTreeCheckpointer()
  save_args = orbax_utils.save_args_from_target(params)
  orbax_checkpointer.save(ckpt_path, params, force=True, save_args=save_args)

  config_path = ckpt_path / config_fname
  config_path.write_text(config.to_json_best_effort())


def load(
    path: Union[str, epath.Path],
):
  """Loads checkpoint."""
  path = epath.Path(path)
  if not path.exists():
    raise ValueError(f'checkpoint path does not exist: {path.as_posix()}')

  logging.info('restoring from checkpoint %s', path.as_posix())

  metadata = ocp.PyTreeCheckpointer().metadata(path)
  restore_args = jax.tree.map(
      lambda _: ocp.RestoreArgs(restore_type=np.ndarray), metadata
  )
  orbax_checkpointer = ocp.PyTreeCheckpointer()
  target = orbax_checkpointer.restore(
      path, ocp.args.PyTreeRestore(restore_args=restore_args), item=None
  )
  target[0] = running_statistics.RunningStatisticsState(**target[0])

  return target


def load_config(
    config_path: Union[str, epath.Path],
) -> config_dict.ConfigDict:
  """Loads config from config path."""
  config_path = epath.Path(config_path)
  if not config_path.exists():
    raise ValueError(f'Config file not found at {config_path.as_posix()}')
  return config_dict.create(**json.loads(config_path.read_text()))



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Brax training types."""

from typing import Any, Mapping, NamedTuple, Tuple, TypeVar, Union

from brax.training.acme.types import NestedArray
import flax
import jax
import jax.numpy as jnp
import numpy as np

# Protocol was introduced into typing in Python >=3.8
# via https://www.python.org/dev/peps/pep-0544/
# Before that, its status was DRAFT and available via typing_extensions
try:
  from typing import Protocol  # pylint:disable=g-import-not-at-top
except ImportError:
  from typing_extensions import Protocol  # pylint:disable=g-import-not-at-top

Params = Any
PRNGKey = jnp.ndarray
Metrics = Mapping[str, jnp.ndarray]
Observation = Union[jnp.ndarray, Mapping[str, jnp.ndarray]]
ObservationSize = Union[int, Mapping[str, Union[Tuple[int, ...], int]]]
Action = jnp.ndarray
Extra = Mapping[str, Any]
PolicyParams = Any
PreprocessorParams = Any
PolicyParams = Tuple[PreprocessorParams, Params]
NetworkType = TypeVar('NetworkType')


class Transition(NamedTuple):
  """Container for a transition."""

  observation: NestedArray
  action: NestedArray
  reward: NestedArray
  discount: NestedArray
  next_observation: NestedArray
  extras: NestedArray = ()  # pytype: disable=annotation-type-mismatch  # jax-ndarray


class Policy(Protocol):

  def __call__(
      self,
      observation: Observation,
      key: PRNGKey,
  ) -> Tuple[Action, Extra]:
    pass


class PreprocessObservationFn(Protocol):

  def __call__(
      self,
      observation: Observation,
      preprocessor_params: PreprocessorParams,
  ) -> jnp.ndarray:
    pass


def identity_observation_preprocessor(
    observation: Observation, preprocessor_params: PreprocessorParams
):
  del preprocessor_params
  return observation


class NetworkFactory(Protocol[NetworkType]):

  def __call__(
      self,
      observation_size: ObservationSize,
      action_size: int,
      preprocess_observations_fn: PreprocessObservationFn = identity_observation_preprocessor,
  ) -> NetworkType:
    pass


@flax.struct.dataclass
class UInt64:
  """Custom 64-bit integer implementation using two 32-bit parts.

  This class implements 64-bit unsigned integer arithmetic using two 32-bit
  parts
  (hi and lo) to work around JAX's limitation of not supporting 64-bit
  unsigned integers directly when jax_enable_x64 is False.
  """

  hi: Union[int, np.ndarray, jax.Array]
  lo: Union[int, np.ndarray, jax.Array]

  def to_numpy(self):
    """Convert UInt64 to numpy uint64."""
    hi_np = np.array(self.hi, dtype=np.uint64)
    lo_np = np.array(self.lo, dtype=np.uint64)
    return (hi_np << np.uint64(32)) | lo_np

  def __post_init__(self):
    """Cast post init."""
    object.__setattr__(self, "hi", jnp.uint32(self.hi))
    object.__setattr__(self, "lo", jnp.uint32(self.lo))

  def __add__(self, other):
    other = _sanitize_uint64_input(other)
    return _add_uint64(self, other)

  def __repr__(self):
    return f"UInt64(hi={self.hi}, lo={self.lo})"

  def __int__(self):
    """Convert UInt64 to Python int."""
    return int(self.to_numpy())


def _sanitize_uint64_input(other: Union[int, np.ndarray, jax.Array, UInt64]):
  """Sanitizes input for UInt64 arithmetic.

  Args:
    other: Input value, either int, np.ndarray or UInt64.

  Returns:
    UInt64 representation of the input.

  Raises:
    NotImplementedError: If the input type is not supported.
  """
  if isinstance(other, (int, np.ndarray, jax.Array)):
    other_lo = other & jnp.array(0xFFFFFFFF, dtype=jnp.uint32)
    other_hi = other >> 32
    other = UInt64(
        hi=jnp.array(other_hi, dtype=jnp.uint32),
        lo=jnp.array(other_lo, dtype=jnp.uint32),
    )
  elif isinstance(other, UInt64):
    pass
  else:
    raise NotImplementedError(f"Cannot perform op on {type(other)} and UInt64.")
  return other


def _add_uint64(a: UInt64, b: UInt64) -> UInt64:
  """Adds two UInt64 numbers together.

  Args:
    a: First UInt64 number
    b: Second UInt64 number

  Returns:
    New UInt64 number representing the sum
  """
  result_lo = a.lo + b.lo
  carry_bit_pattern = (a.lo & b.lo) | ((a.lo | b.lo) & (~result_lo))
  carry = carry_bit_pattern >> 31
  result_hi = a.hi + b.hi + carry
  return UInt64(hi=result_hi, lo=result_lo)



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test PPO checkpointing."""

import functools

from absl import flags
from absl.testing import absltest
from brax.training.acme import running_statistics
from brax.training.agents.sac import checkpoint
from brax.training.agents.sac import losses as sac_losses
from brax.training.agents.sac import networks as sac_networks
from etils import epath
import jax
from jax import numpy as jp


class CheckpointTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    flags.FLAGS.mark_as_parsed()

  def test_sac_params_config(self):
    network_factory = functools.partial(
        sac_networks.make_sac_networks,
        hidden_layer_sizes=(16, 21, 13),
    )
    config = checkpoint.network_config(
        action_size=3,
        observation_size=1,
        normalize_observations=True,
        network_factory=network_factory,
    )
    self.assertEqual(
        config.network_factory_kwargs.to_dict()["hidden_layer_sizes"],
        (16, 21, 13),
    )
    self.assertEqual(config.action_size, 3)
    self.assertEqual(config.observation_size, 1)

  def test_save_and_load_checkpoint(self):
    path = self.create_tempdir("test")
    network_factory = functools.partial(
        sac_networks.make_sac_networks,
        hidden_layer_sizes=(16, 21, 13),
    )
    config = checkpoint.network_config(
        observation_size=1,
        action_size=3,
        normalize_observations=True,
        network_factory=network_factory,
    )

    # Generate network params for saving a dummy checkpoint.
    normalize = lambda x, y: x
    if config.normalize_observations:
      normalize = running_statistics.normalize
    sac_network = network_factory(
        config.observation_size,
        config.action_size,
        preprocess_observations_fn=normalize,
        **config.network_factory_kwargs,
    )
    dummy_key = jax.random.PRNGKey(0)
    normalizer_params = running_statistics.init_state(
        jax.tree_util.tree_map(jp.zeros, config.observation_size)
    )
    params = (normalizer_params, sac_network.policy_network.init(dummy_key))

    # Save and load a checkpoint.
    checkpoint.save(
        path.full_path,
        step=1,
        params=params,
        config=config,
    )

    policy_fn = checkpoint.load_policy(
        epath.Path(path.full_path) / "000000000001",
    )
    out = policy_fn(jp.zeros(1), jax.random.PRNGKey(0))
    self.assertEqual(out[0].shape, (3,))


if __name__ == "__main__":
  absltest.main()



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Network definitions."""

import dataclasses
import functools
from typing import Any, Callable, Literal, Mapping, Sequence, Tuple
import warnings

from brax.training import types
from brax.training.acme import running_statistics
from brax.training.spectral_norm import SNDense
from flax import linen
import jax
import jax.numpy as jnp

ActivationFn = Callable[[jnp.ndarray], jnp.ndarray]
Initializer = Callable[..., Any]


@dataclasses.dataclass
class FeedForwardNetwork:
  init: Callable[..., Any]
  apply: Callable[..., Any]


class MLPHead(linen.Module):
  """MLP over pre-processed latent vectors.

  For an example usage, see the Aloha sim2real code on
  https://github.com/google-deepmind/mujoco_playground.
  """

  layer_sizes: Sequence[int]
  activation: ActivationFn = linen.relu
  kernel_init: Initializer = jax.nn.initializers.lecun_uniform()
  activate_final: bool = False
  bias: bool = True
  layer_norm: bool = False
  state_key: str = 'proprio'
  latent_key_prefix: str = 'latent_'  # Assume followed by integer index.
  latent_head_size: int = 64

  @linen.compact
  def __call__(self, data: Mapping[str, jax.Array]):
    latents = []
    latent_keys = sorted(
        [k for k in data.keys() if k.startswith(self.latent_key_prefix)],
        key=lambda x: int(x.split('_')[-1]),
    )
    assert len(latent_keys) > 0, 'No latent keys found'
    for key in latent_keys:
      latents.append(data[key])
    hidden = [
        self.activation(linen.Dense(self.latent_head_size)(latent))
        for latent in latents
    ]
    if self.state_key:
      hidden.append(data[self.state_key])
    hidden = jnp.concatenate(hidden, axis=-1)
    return MLP(
        layer_sizes=self.layer_sizes,
        activation=self.activation,
        kernel_init=self.kernel_init,
        activate_final=self.activate_final,
        layer_norm=self.layer_norm,
    )(hidden)


class MLP(linen.Module):
  """MLP module."""

  layer_sizes: Sequence[int]
  activation: ActivationFn = linen.relu
  kernel_init: Initializer = jax.nn.initializers.lecun_uniform()
  activate_final: bool = False
  bias: bool = True
  layer_norm: bool = False

  @linen.compact
  def __call__(self, data: jnp.ndarray):
    hidden = data
    for i, hidden_size in enumerate(self.layer_sizes):
      hidden = linen.Dense(
          hidden_size,
          name=f'hidden_{i}',
          kernel_init=self.kernel_init,
          use_bias=self.bias,
      )(hidden)
      if i != len(self.layer_sizes) - 1 or self.activate_final:
        hidden = self.activation(hidden)
        if self.layer_norm:
          hidden = linen.LayerNorm()(hidden)
    return hidden


class SNMLP(linen.Module):
  """MLP module with Spectral Normalization."""

  layer_sizes: Sequence[int]
  activation: ActivationFn = linen.relu
  kernel_init: Initializer = jax.nn.initializers.lecun_uniform()
  activate_final: bool = False
  bias: bool = True

  @linen.compact
  def __call__(self, data: jnp.ndarray):
    hidden = data
    for i, hidden_size in enumerate(self.layer_sizes):
      hidden = SNDense(
          hidden_size,
          name=f'hidden_{i}',
          kernel_init=self.kernel_init,
          use_bias=self.bias,
      )(hidden)
      if i != len(self.layer_sizes) - 1 or self.activate_final:
        hidden = self.activation(hidden)
    return hidden


class CNN(linen.Module):
  """CNN module. Inputs are expected in Batch * HWC format."""

  num_filters: Sequence[int]
  kernel_sizes: Sequence[Tuple]
  strides: Sequence[Tuple]
  activation: ActivationFn = linen.relu
  use_bias: bool = True

  @linen.compact
  def __call__(self, data: jnp.ndarray):
    hidden = data
    for i, (num_filter, kernel_size, stride) in enumerate(
        zip(self.num_filters, self.kernel_sizes, self.strides)
    ):
      hidden = linen.Conv(
          num_filter,
          kernel_size=kernel_size,
          strides=stride,
          use_bias=self.use_bias,
      )(hidden)

      hidden = self.activation(hidden)
    return hidden


class VisionMLP(linen.Module):
  """Applies a CNN backbone then an MLP.

  The CNN architecture originates from the paper:
  "Human-level control through deep reinforcement learning",
  Nature 518, no. 7540 (2015): 529-533
  """

  layer_sizes: Sequence[int]
  activation: ActivationFn = linen.relu
  kernel_init: Initializer = jax.nn.initializers.lecun_uniform()
  activate_final: bool = False
  layer_norm: bool = False
  normalise_channels: bool = False
  state_obs_key: str = ''
  policy_head: bool = True  # = False is useful for frozen encoders.

  @linen.compact
  def __call__(self, data: dict):
    pixels_hidden = {k: v for k, v in data.items() if k.startswith('pixels/')}
    if self.normalise_channels:
      # Calculates shared statistics over an entire 2D image.
      image_layernorm = functools.partial(
          linen.LayerNorm,
          use_bias=False,
          use_scale=False,
          reduction_axes=(-1, -2),
      )

      def ln_per_chan(v: jax.Array):
        normalised = [
            image_layernorm()(v[..., chan]) for chan in range(v.shape[-1])
        ]
        return jnp.stack(normalised, axis=-1)

      pixels_hidden = jax.tree.map(ln_per_chan, pixels_hidden)

    natureCNN = functools.partial(
        CNN,
        num_filters=[32, 64, 64],
        kernel_sizes=[(8, 8), (4, 4), (3, 3)],
        strides=[(4, 4), (2, 2), (1, 1)],
        activation=linen.relu,
        use_bias=False,
    )
    cnn_outs = [natureCNN()(pixels_hidden[key]) for key in pixels_hidden]
    cnn_outs = [jnp.mean(cnn_out, axis=(-2, -3)) for cnn_out in cnn_outs]
    if not self.policy_head:
      return jnp.concatenate(cnn_outs, axis=-1)
    if self.state_obs_key:
      cnn_outs.append(
          data[self.state_obs_key]
      )  # TODO: Try with dedicated state network

    hidden = jnp.concatenate(cnn_outs, axis=-1)
    return MLP(
        layer_sizes=self.layer_sizes,
        activation=self.activation,
        kernel_init=self.kernel_init,
        activate_final=self.activate_final,
        layer_norm=self.layer_norm,
    )(hidden)


def _get_obs_state_size(obs_size: types.ObservationSize, obs_key: str) -> int:
  obs_size = obs_size[obs_key] if isinstance(obs_size, Mapping) else obs_size
  return jax.tree_util.tree_flatten(obs_size)[0][-1]


class Param(linen.Module):
  """Scalar parameter module."""

  init_value: float = 0.0
  size: int = 1

  @linen.compact
  def __call__(self):
    return self.param(
        'value', init_fn=lambda keys: jnp.full((self.size,), self.init_value)
    )


class LogParam(linen.Module):
  """Scalar parameter module with log scale."""

  init_value: float = 1.0
  size: int = 1

  @linen.compact
  def __call__(self):
    log_value = self.param(
        'log_value',
        init_fn=lambda key: jnp.full((self.size,), jnp.log(self.init_value)),
    )
    return jnp.exp(log_value)


class PolicyModuleWithStd(linen.Module):
  """Policy module with learnable mean and standard deviation."""

  param_size: int
  hidden_layer_sizes: Sequence[int]
  activation: ActivationFn
  kernel_init: jax.nn.initializers.Initializer
  layer_norm: bool
  noise_std_type: Literal['scalar', 'log']
  init_noise_std: float
  state_dependent_std: bool = False

  @linen.compact
  def __call__(self, obs):
    if self.noise_std_type not in ['scalar', 'log']:
      raise ValueError(
          f'Unsupported noise std type: {self.noise_std_type}. Must be one of'
          ' "scalar" or "log".'
      )

    outputs = MLP(
        layer_sizes=list(self.hidden_layer_sizes),
        activation=self.activation,
        kernel_init=self.kernel_init,
        layer_norm=self.layer_norm,
        activate_final=True,
    )(obs)

    mean_params = linen.Dense(
        self.param_size, kernel_init=self.kernel_init
    )(outputs)

    if self.state_dependent_std:
      log_std_output = linen.Dense(
          self.param_size, kernel_init=self.kernel_init
      )(outputs)
      if self.noise_std_type == 'log':
        std_params = jnp.exp(log_std_output)
      else:
        std_params = log_std_output

    else:
      if self.noise_std_type == 'scalar':
        std_module = Param(
            self.init_noise_std, size=self.param_size, name='std_param'
        )
      else:
        std_module = LogParam(
            self.init_noise_std, size=self.param_size, name='std_logparam'
        )
      std_params = std_module()

    return mean_params, std_params


def make_policy_network(
    param_size: int,
    obs_size: types.ObservationSize,
    preprocess_observations_fn: types.PreprocessObservationFn = types.identity_observation_preprocessor,
    hidden_layer_sizes: Sequence[int] = (256, 256),
    activation: ActivationFn = linen.relu,
    kernel_init: Initializer = jax.nn.initializers.lecun_uniform(),
    layer_norm: bool = False,
    obs_key: str = 'state',
    distribution_type: Literal['normal', 'tanh_normal'] = 'tanh_normal',
    noise_std_type: Literal['scalar', 'log'] = 'scalar',
    init_noise_std: float = 1.0,
    state_dependent_std: bool = False,
) -> FeedForwardNetwork:
  """Creates a policy network."""
  if distribution_type == 'tanh_normal':
    policy_module = MLP(
        layer_sizes=list(hidden_layer_sizes) + [param_size],
        activation=activation,
        kernel_init=kernel_init,
        layer_norm=layer_norm,
    )
  elif distribution_type == 'normal':
    policy_module = PolicyModuleWithStd(
        param_size=param_size,
        hidden_layer_sizes=hidden_layer_sizes,
        activation=activation,
        kernel_init=kernel_init,
        layer_norm=layer_norm,
        noise_std_type=noise_std_type,
        init_noise_std=init_noise_std,
        state_dependent_std=state_dependent_std,
    )
  else:
    raise ValueError(
        f'Unsupported distribution type: {distribution_type}. Must be one'
        ' of "normal" or "tanh_normal".'
    )

  def apply(processor_params, policy_params, obs):
    if isinstance(obs, Mapping):
      obs = preprocess_observations_fn(
          obs[obs_key], normalizer_select(processor_params, obs_key)
      )
    else:
      obs = preprocess_observations_fn(obs, processor_params)
    return policy_module.apply(policy_params, obs)

  obs_size = _get_obs_state_size(obs_size, obs_key)
  dummy_obs = jnp.zeros((1, obs_size))

  def init(key):
    policy_module_params = policy_module.init(key, dummy_obs)
    return policy_module_params

  return FeedForwardNetwork(init=init, apply=apply)


def make_value_network(
    obs_size: types.ObservationSize,
    preprocess_observations_fn: types.PreprocessObservationFn = types.identity_observation_preprocessor,
    hidden_layer_sizes: Sequence[int] = (256, 256),
    activation: ActivationFn = linen.relu,
    obs_key: str = 'state',
) -> FeedForwardNetwork:
  """Creates a value network."""
  value_module = MLP(
      layer_sizes=list(hidden_layer_sizes) + [1],
      activation=activation,
      kernel_init=jax.nn.initializers.lecun_uniform(),
  )

  def apply(processor_params, value_params, obs):
    if isinstance(obs, Mapping):
      obs = preprocess_observations_fn(
          obs[obs_key], normalizer_select(processor_params, obs_key)
      )
    else:
      obs = preprocess_observations_fn(obs, processor_params)
    return jnp.squeeze(value_module.apply(value_params, obs), axis=-1)

  obs_size = _get_obs_state_size(obs_size, obs_key)
  dummy_obs = jnp.zeros((1, obs_size))
  return FeedForwardNetwork(
      init=lambda key: value_module.init(key, dummy_obs), apply=apply
  )


def make_q_network(
    obs_size: types.ObservationSize,
    action_size: int,
    preprocess_observations_fn: types.PreprocessObservationFn = types.identity_observation_preprocessor,
    hidden_layer_sizes: Sequence[int] = (256, 256),
    activation: ActivationFn = linen.relu,
    n_critics: int = 2,
    layer_norm: bool = False,
) -> FeedForwardNetwork:
  """Creates a value network."""

  class QModule(linen.Module):
    """Q Module."""

    n_critics: int

    @linen.compact
    def __call__(self, obs: jnp.ndarray, actions: jnp.ndarray):
      hidden = jnp.concatenate([obs, actions], axis=-1)
      res = []
      for _ in range(self.n_critics):
        q = MLP(
            layer_sizes=list(hidden_layer_sizes) + [1],
            activation=activation,
            kernel_init=jax.nn.initializers.lecun_uniform(),
            layer_norm=layer_norm,
        )(hidden)
        res.append(q)
      return jnp.concatenate(res, axis=-1)

  q_module = QModule(n_critics=n_critics)

  def apply(processor_params, q_params, obs, actions):
    obs = preprocess_observations_fn(obs, processor_params)
    return q_module.apply(q_params, obs, actions)

  dummy_obs = jnp.zeros((1, obs_size))
  dummy_action = jnp.zeros((1, action_size))
  return FeedForwardNetwork(
      init=lambda key: q_module.init(key, dummy_obs, dummy_action), apply=apply
  )


def make_model(
    layer_sizes: Sequence[int],
    obs_size: int,
    activation: Callable[[jnp.ndarray], jnp.ndarray] = linen.swish,
    spectral_norm: bool = False,
) -> FeedForwardNetwork:
  """Creates a model.

  Args:
    layer_sizes: layers
    obs_size: size of an observation
    activation: activation
    spectral_norm: whether to use a spectral normalization (default: False).

  Returns:
    a model
  """
  warnings.warn(
      'make_model is deprecated, use make_{policy|q|value}_network instead.'
  )
  dummy_obs = jnp.zeros((1, obs_size))
  if spectral_norm:
    module = SNMLP(layer_sizes=layer_sizes, activation=activation)
    model = FeedForwardNetwork(
        init=lambda rng1, rng2: module.init(
            {'params': rng1, 'sing_vec': rng2}, dummy_obs
        ),
        apply=module.apply,
    )
  else:
    module = MLP(layer_sizes=layer_sizes, activation=activation)
    model = FeedForwardNetwork(
        init=lambda rng: module.init(rng, dummy_obs), apply=module.apply
    )
  return model


def make_models(
    policy_params_size: int, obs_size: int
) -> Tuple[FeedForwardNetwork, FeedForwardNetwork]:
  """Creates models for policy and value functions.

  Args:
    policy_params_size: number of params that a policy network should generate
    obs_size: size of an observation

  Returns:
    a model for policy and a model for value function
  """
  warnings.warn(
      'make_models is deprecated, use make_{policy|q|value}_network instead.'
  )
  policy_model = make_model([32, 32, 32, 32, policy_params_size], obs_size)
  value_model = make_model([256, 256, 256, 256, 256, 1], obs_size)
  return policy_model, value_model


def normalizer_select(
    processor_params: running_statistics.RunningStatisticsState, obs_key: str
) -> running_statistics.RunningStatisticsState:
  return running_statistics.RunningStatisticsState(
      count=processor_params.count,
      mean=processor_params.mean[obs_key],
      summed_variance=processor_params.summed_variance[obs_key],
      std=processor_params.std[obs_key],
  )


def make_policy_network_vision(
    observation_size: Mapping[str, Tuple[int, ...]],
    output_size: int,
    preprocess_observations_fn: types.PreprocessObservationFn = types.identity_observation_preprocessor,
    hidden_layer_sizes: Sequence[int] = [256, 256],
    activation: ActivationFn = linen.swish,
    kernel_init: Initializer = jax.nn.initializers.lecun_uniform(),
    layer_norm: bool = False,
    state_obs_key: str = '',
    normalise_channels: bool = False,
) -> FeedForwardNetwork:
  """Creates a policy network for vision inputs."""
  module = VisionMLP(
      layer_sizes=list(hidden_layer_sizes) + [output_size],
      activation=activation,
      kernel_init=kernel_init,
      layer_norm=layer_norm,
      normalise_channels=normalise_channels,
      state_obs_key=state_obs_key,
  )

  def apply(processor_params, policy_params, obs):
    if state_obs_key:
      state_obs = preprocess_observations_fn(
          obs[state_obs_key], normalizer_select(processor_params, state_obs_key)
      )
      obs = {**obs, state_obs_key: state_obs}
    return module.apply(policy_params, obs)

  dummy_obs = {
      key: jnp.zeros((1,) + shape) for key, shape in observation_size.items()
  }
  return FeedForwardNetwork(
      init=lambda key: module.init(key, dummy_obs), apply=apply
  )


def make_value_network_vision(
    observation_size: Mapping[str, Tuple[int, ...]],
    preprocess_observations_fn: types.PreprocessObservationFn = types.identity_observation_preprocessor,
    hidden_layer_sizes: Sequence[int] = [256, 256],
    activation: ActivationFn = linen.swish,
    kernel_init: Initializer = jax.nn.initializers.lecun_uniform(),
    state_obs_key: str = '',
    normalise_channels: bool = False,
) -> FeedForwardNetwork:
  """Creates a value network for vision inputs."""
  value_module = VisionMLP(
      layer_sizes=list(hidden_layer_sizes) + [1],
      activation=activation,
      kernel_init=kernel_init,
      normalise_channels=normalise_channels,
      state_obs_key=state_obs_key,
  )

  def apply(processor_params, policy_params, obs):
    if state_obs_key:
      state_obs = preprocess_observations_fn(
          obs[state_obs_key], normalizer_select(processor_params, state_obs_key)
      )
      obs = {**obs, state_obs_key: state_obs}
    return jnp.squeeze(value_module.apply(policy_params, obs), axis=-1)

  dummy_obs = {
      key: jnp.zeros((1,) + shape) for key, shape in observation_size.items()
  }
  return FeedForwardNetwork(
      init=lambda key: value_module.init(key, dummy_obs), apply=apply
  )


def make_policy_network_latents(
    param_size: int,
    observation_size: types.ObservationSize,
    preprocess_observations_fn: types.PreprocessObservationFn = types.identity_observation_preprocessor,
    hidden_layer_sizes: Sequence[int] = (256, 256),
    activation: ActivationFn = linen.relu,
    kernel_init: Initializer = jax.nn.initializers.lecun_uniform(),
    layer_norm: bool = False,
    obs_key: str = 'state',
) -> FeedForwardNetwork:
  """Creates a policy network. Wraps MLPLatents.

  Has same API as make_policy_network. Used when the env observation has image
  latents rather than pixels.

  Returns:
    A FeedForwardNetwork that takes observations and returns policy
    parameters.
  """
  module = MLPHead(
      layer_sizes=list(hidden_layer_sizes) + [param_size],
      activation=activation,
      kernel_init=kernel_init,
      layer_norm=layer_norm,
      state_key=obs_key,
  )

  def apply(processor_params, policy_params, obs):
    if obs_key:
      state_obs = preprocess_observations_fn(
          obs[obs_key], normalizer_select(processor_params, obs_key)
      )
      obs = {**obs, obs_key: state_obs}
    return module.apply(policy_params, obs)

  if not isinstance(observation_size, Mapping):
    raise NotImplementedError(
        'make_policy_network_latents only implemented for dictionary'
        ' observations'
    )

  dummy_obs = {
      key: jnp.zeros((1,) + shape) for key, shape in observation_size.items()
  }
  return FeedForwardNetwork(
      init=lambda key: module.init(key, dummy_obs), apply=apply
  )



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SAC networks."""

from typing import Literal, Sequence, Tuple

from brax.training import distribution
from brax.training import networks
from brax.training import types
from brax.training.types import PRNGKey
import flax
from flax import linen


@flax.struct.dataclass
class SACNetworks:
  policy_network: networks.FeedForwardNetwork
  q_network: networks.FeedForwardNetwork
  parametric_action_distribution: distribution.ParametricDistribution


def make_inference_fn(sac_networks: SACNetworks):
  """Creates params and inference function for the SAC agent."""

  def make_policy(
      params: types.PolicyParams, deterministic: bool = False
  ) -> types.Policy:

    def policy(
        observations: types.Observation, key_sample: PRNGKey
    ) -> Tuple[types.Action, types.Extra]:
      logits = sac_networks.policy_network.apply(*params, observations)
      if deterministic:
        return sac_networks.parametric_action_distribution.mode(logits), {}
      return (
          sac_networks.parametric_action_distribution.sample(
              logits, key_sample
          ),
          {},
      )

    return policy

  return make_policy


def make_sac_networks(
    observation_size: int,
    action_size: int,
    preprocess_observations_fn: types.PreprocessObservationFn = types.identity_observation_preprocessor,
    hidden_layer_sizes: Sequence[int] = (256, 256),
    activation: networks.ActivationFn = linen.relu,
    policy_network_layer_norm: bool = False,
    q_network_layer_norm: bool = False,
    distribution_type: Literal['normal', 'tanh_normal'] = 'tanh_normal',
    noise_std_type: Literal['scalar', 'log'] = 'scalar',
    init_noise_std: float = 1.0,
    state_dependent_std: bool = False,
) -> SACNetworks:
  """Make SAC networks."""
  parametric_action_distribution: distribution.ParametricDistribution
  if distribution_type == 'normal':
    parametric_action_distribution = distribution.NormalDistribution(
        event_size=action_size
    )
  elif distribution_type == 'tanh_normal':
    parametric_action_distribution = distribution.NormalTanhDistribution(
        event_size=action_size
    )
  else:
    raise ValueError(
        f'Unsupported distribution type: {distribution_type}. Must be one'
        ' of "normal" or "tanh_normal".'
    )
  policy_network = networks.make_policy_network(
      parametric_action_distribution.param_size,
      observation_size,
      preprocess_observations_fn=preprocess_observations_fn,
      hidden_layer_sizes=hidden_layer_sizes,
      activation=activation,
      layer_norm=policy_network_layer_norm,
      distribution_type=distribution_type,
      noise_std_type=noise_std_type,
      init_noise_std=init_noise_std,
      state_dependent_std=state_dependent_std,
  )
  q_network = networks.make_q_network(
      observation_size,
      action_size,
      preprocess_observations_fn=preprocess_observations_fn,
      hidden_layer_sizes=hidden_layer_sizes,
      activation=activation,
      layer_norm=q_network_layer_norm,
  )
  return SACNetworks(
      policy_network=policy_network,
      q_network=q_network,
      parametric_action_distribution=parametric_action_distribution,
  )



# Copyright 2025 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BC tests."""

import functools
import pickle

from absl.testing import absltest
from absl.testing import parameterized
from brax import envs
from brax.training.acme import running_statistics
from brax.training.agents.bc import networks as bc_networks
from brax.training.agents.bc import train as bc
import jax


class BCTest(parameterized.TestCase):
  """Tests for BC module."""

  def _create_teacher_policy(self, num_envs):
    """Helper to create a dummy teacher policy."""

    def teacher_policy(obs, rng):
      return (
          jax.numpy.ones((num_envs, 1)),
          {'loc': 1 * jax.numpy.ones((num_envs, 1))},
      )

    return teacher_policy

  def testTrain(self):
    """Test BC with a simple env."""
    teacher_policy = self._create_teacher_policy(num_envs=50)
    fast = envs.get_environment('fast', obs_mode='dict_state')
    fast = envs.training.wrap(fast, episode_length=128, action_repeat=1)
    network_factory = functools.partial(
        bc_networks.make_bc_networks,
        policy_hidden_layer_sizes=(32,),
        policy_obs_key='state',
        vision=False,
    )
    _, _, metrics = bc.train(
        env=fast,
        teacher_inference_fn=teacher_policy,
        demo_length=6,
        num_envs=50,
        eval_length=128,
        num_evals=3,
        num_eval_envs=10,
        dagger_steps=20,
        batch_size=50,
        learning_rate=3e-3,
        epochs=4,
        network_factory=network_factory,
        reset=False,
    )
    self.assertGreater(metrics['eval/episode_reward'], 135)

  def testNetworkEncoding(self):
    """Test network encoding and decoding."""
    teacher_policy = self._create_teacher_policy(num_envs=50)
    fast = envs.get_environment('fast', obs_mode='dict_state')
    w_fast = envs.training.wrap(fast, episode_length=128, action_repeat=1)

    # Train a BC agent to get parameters
    network_factory = functools.partial(
        bc_networks.make_bc_networks,
        policy_hidden_layer_sizes=(32,),
        policy_obs_key='state',
        vision=False,
    )

    original_inference, params, _ = bc.train(
        env=w_fast,
        teacher_inference_fn=teacher_policy,
        demo_length=6,
        num_envs=50,
        eval_length=128,
        num_evals=1,
        dagger_steps=2,
        batch_size=50,
        learning_rate=3e-3,
        epochs=2,
        tanh_squash=True,
        network_factory=network_factory,
        reset=False,
    )

    # Create a new inference function with the same network structure
    normalize_fn = running_statistics.normalize

    bc_network = network_factory(
        fast.observation_size, fast.action_size, normalize_fn
    )
    make_inference = bc_networks.make_inference_fn(bc_network)

    # Test serialization and deserialization
    byte_encoding = pickle.dumps(params)
    decoded_params = pickle.loads(byte_encoding)

    # Compute one action with both the original and the reconstructed parameters
    state = fast.reset(jax.random.PRNGKey(0))
    original_action = original_inference(decoded_params)(
        state.obs, jax.random.PRNGKey(0)
    )[0]
    action = make_inference(decoded_params, tanh_squash=True)(
        state.obs, jax.random.PRNGKey(0)
    )[0]

    # Verify that the actions are the same
    self.assertSequenceEqual(original_action, action)
    fast.step(state, action)

  def testPixelsBC(self):
    """Test BC with pixel observations."""
    teacher_policy = self._create_teacher_policy(num_envs=50)
    fast = envs.get_environment('fast', obs_mode='dict_latent_state')
    fast = envs.training.wrap(fast, episode_length=128, action_repeat=1)
    network_factory = functools.partial(
        bc_networks.make_bc_networks,
        policy_hidden_layer_sizes=(32,),
        policy_obs_key='state',
        latent_vision=True,
    )
    _, _, metrics = bc.train(
        env=fast,
        teacher_inference_fn=teacher_policy,
        demo_length=6,
        num_envs=50,
        eval_length=128,
        num_evals=3,
        num_eval_envs=10,
        dagger_steps=20,
        batch_size=50,
        learning_rate=3e-3,
        epochs=4,
        network_factory=network_factory,
        reset=False,
        augment_pixels=True,
    )
    self.assertGreater(metrics['eval/episode_reward'], 135)


if __name__ == '__main__':
  absltest.main()
